from ase import Atoms
from gpaw import GPAW, FermiDirac
from gpaw import PW
from gpaw.bztools import find_high_symmetry_monkhorst_pack
from gpaw.response.df import DielectricFunction
from gpaw.test import findpeak
import numpy as np
from scipy.ndimage import gaussian_filter1d


def test_point_tetra_match(in_tmp_dir):
    
    gs_file = 'gs.gpw'
    response_file = 'gsresponse.gpw'
    
    # Create graphene lattice
    a = 2.5
    atoms = Atoms(
        symbols='C2', positions=[(0.5 * a, -np.sqrt(3) / 6 * a, 0.0),
                                 (0.5 * a, np.sqrt(3) / 6 * a, 0.0)],
        cell=[(0.5 * a, -0.5 * 3**0.5 * a, 0),
              (0.5 * a, 0.5 * 3**0.5 * a, 0),
              (0.0, 0.0, 3.2)],
        pbc=[True, True, False])

    atoms.center(axis=2)

    calc = GPAW(mode=PW(400),
                kpts={'density': 10.0, 'gamma': True},
                occupations=FermiDirac(0.1))

    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write(gs_file)

    density = 15
    kpts = find_high_symmetry_monkhorst_pack(gs_file, density=density)
    responseGS = GPAW(gs_file).fixed_density(
        kpts=kpts,
        parallel={'band': 1},
        nbands=20,
        occupations=FermiDirac(0.001),
        convergence={'bands': 10})
    responseGS.write(response_file, 'all')

    # DF with tetra integration
    df = DielectricFunction(response_file,
                            eta=25e-3,
                            rate='eta',
                            frequencies={'type': 'nonlinear',
                                         'domega0': 0.1},
                            integrationmode='tetrahedron integration')
    df1_tetra, df2_tetra = df.get_dielectric_function(q_c=[0, 0, 0])

    # DF with point integration
    df = DielectricFunction(response_file,
                            frequencies={'type': 'nonlinear',
                                         'domega0': 0.1},
                            eta=25e-3,
                            rate='eta')
    df1_point, df2_point = df.get_dielectric_function(q_c=[0, 0, 0])

    omega = df.get_frequencies()
    df2_tetra = np.imag(df2_tetra)
    df2_point = np.imag(df2_point)
    
    # Do not use frequencies near the w=0 singularity
    slicer = [(freq >= 1.5) and (freq <= 20) for freq in omega]
    # Convolve with Gaussian to smoothen the curve
    sigma = 1.9
    df2_gauss = gaussian_filter1d(df2_point[slicer], sigma)
    
    rms_diff_tetra_point = np.sqrt(np.sum((df2_tetra[slicer]
                                   - df2_point[slicer])**2) / np.sum(slicer))
    rms_diff_tetra_gauss = np.sqrt(np.sum((df2_tetra[slicer]
                                   - df2_gauss)**2) / np.sum(slicer))

    assert rms_diff_tetra_point < 1.35
    assert rms_diff_tetra_gauss < 1.10
    assert rms_diff_tetra_point * 0.80 > rms_diff_tetra_gauss
    
    freq1, amp1 = findpeak(omega[slicer], df2_tetra[slicer])
    freq2, amp2 = findpeak(omega[slicer], df2_gauss)

    assert freq1 - freq2 < 0.1

    # # Plot the figures
    # import matplotlib.pyplot as plt
    # plt.figure(figsize=(6, 6))
    # plt.plot(omega[slicer], df2_tetra[slicer], label='Tetrahedron')
    # plt.plot(omega[slicer], df2_point[slicer], label='Point sampling')
    # plt.plot(omega[slicer], df2_gauss, label='Gaussian convolution')
    # plt.xlabel('Frequency (eV)')
    # plt.ylabel('$\\mathrm{Im}\\varepsilon$')
    # plt.xlim(0, 6)
    # plt.ylim(0, 20)
    # plt.legend()
    # plt.tight_layout()
    # plt.savefig('graphene_eps.png', dpi=300)



import pytest
from gpaw.xc.fxc import FXCCorrelation


@pytest.fixture
def ni_gpw(gpw_files, scalapack):
    return gpw_files['ni_pw_kpts333']


@pytest.mark.rpa
@pytest.mark.response
@pytest.mark.parametrize('params, ref_energy', [
    (dict(xc='RPA'), -7.827),
    (dict(xc='rALDA', unit_cells=[2, 1, 1]), -7.501),
    (dict(xc='rAPBE', unit_cells=[2, 1, 1]), -7.456),
])
def test_Ni(in_tmp_dir, ni_gpw, params, ref_energy):
    fxc = FXCCorrelation(ni_gpw, nfrequencies=8, skip_gamma=True,
                         ecut=[50], **params)
    E_fxc = fxc.calculate()
    assert E_fxc == pytest.approx(ref_energy, abs=0.01)



"""
Exponential Transformation Direct Minimization
"""

from gpaw.xc import xc_string_to_dict
from gpaw.directmin.sd_etdm import SteepestDescent, FRcg, QuickMin, LBFGS, \
    LBFGS_P, LSR1P, ModeFollowing
from gpaw.directmin.ls_etdm import MaxStep, StrongWolfeConditions, Parabola


def search_direction(method, etdm=None, pd=None):
    if isinstance(method, str):
        method = xc_string_to_dict(method)

    if isinstance(method, dict):
        kwargs = method.copy()
        names = kwargs.pop('name').replace('-', '').lower().split('_')
        concave_step_length = 0.1
        if 'concave_step_length' in kwargs.keys():
            concave_step_length = kwargs.pop('concave_step_length')

        searchdir = {'sd': SteepestDescent,
                     'frcg': FRcg,
                     'quickmin': QuickMin,
                     'lbfgs': LBFGS,
                     'lbfgsp': LBFGS_P,
                     'lsr1p': LSR1P
                     }[names[0]](**kwargs)

        if len(names) == 2:
            if names[1] == 'gmf':
                pd['gmf'] = True
                searchdir = ModeFollowing(partial_diagonalizer(pd, etdm),
                                          searchdir, concave_step_length)

        return searchdir
    else:
        raise ValueError('Check keyword for search direction!')


def line_search_algorithm(method, objective_function, searchdir_algo):
    if isinstance(method, str):
        method = xc_string_to_dict(method)

    if isinstance(method, dict):
        kwargs = method.copy()
        name = kwargs.pop('name').replace('-', '').lower()
        if name == 'swcawc':
            # for swc-awc we need to know
            # what search. dir. algo is used
            if 'searchdirtype' not in kwargs:
                kwargs['searchdirtype'] = searchdir_algo.type

        ls_algo = {'maxstep': MaxStep,
                   'parabola': Parabola,
                   'swcawc': StrongWolfeConditions
                   }[name](objective_function, **kwargs)

        return ls_algo
    else:
        raise ValueError('Check keyword for line search!')


def partial_diagonalizer(method, domom):
    from gpaw.directmin.derivatives import Davidson
    if isinstance(method, str):
        method = xc_string_to_dict(method)

    if isinstance(method, dict):
        kwargs = method.copy()
        name = kwargs.pop('name')
        if name == 'Davidson':
            return Davidson(domom, **kwargs)
        else:
            raise ValueError('Check keyword for partial diagonalizer!')



import hashlib
from typing import Dict

import numpy as np
from ase.data import atomic_numbers

from gpaw.utilities import pack2
from gpaw.atom.radialgd import AERadialGridDescriptor
from gpaw.atom.configurations import configurations
from gpaw.pseudopotential import PseudoPotential, get_radial_hartree_energy


setups: Dict[str, 'HGHParameterSet'] = {}  # Filled out during parsing below
sc_setups: Dict[str, 'HGHParameterSet'] = {}  # Semicore


# Tabulated values of Gamma(m + 1/2)
half_integer_gamma = [np.sqrt(np.pi)]
for m in range(20):
    half_integer_gamma.append(half_integer_gamma[m] * (m + 0.5))


class HGHSetupData:
    """Setup-compatible class implementing HGH pseudopotential.

    To the PAW code this will appear as a legit PAW setup, but is
    in fact considerably simpler.  In particular, all-electron and
    pseudo partial waves are all zero, and compensation charges do not
    depend on environment.

    A HGH setup has the following form::

                  ----
                   \
      V = Vlocal +  )  | p  > h   < p |
                   /      i    ij    j
                  ----
                   ij

    Vlocal contains a short-range term which is Gaussian-shaped and
    implemented as vbar of a PAW setup, along with a long-range term
    which goes like 1/r and is implemented in terms of a compensation
    charge.

    The non-local part contains KB projector functions which are
    essentially similar to those in PAW, while h_ij are constants.
    h_ij are provided by setting the K_p variable of the normal
    setup.

    Most other properties of a PAW setup do not exist for HGH setups, for
    which reason they are generally set to zero:

    * All-electron partial waves: always zero
    * Pseudo partial waves: always zero
    * Projectors: HGH projectors
    * Zero potential (vbar): Gaussian times polynomial
    * Compensation charges: One Gaussian-shaped spherically symmetric charge
    * All-electron core density: Delta function corresponding to core electron
      charge
    * Pseudo core density: always zero
    * Pseudo valence density: always zero
    * PS/AE Kinetic energy density: always zero
    * The mysterious constants K_p of a setup correspond to h_ij.

    Note that since the pseudo partial waves are set to zero,
    initialization of atomic orbitals requires loading a custom basis
    set.

    Absolute energies become numerically large since no atomic
    reference is subtracted.
    """
    def __init__(self, hghdata):
        if isinstance(hghdata, str):
            symbol = hghdata
            if symbol.endswith('.sc'):
                hghdata = sc_setups[symbol[:-3]]
            else:
                hghdata = setups[symbol]
        self.hghdata = hghdata

        chemsymbol = hghdata.symbol
        if '.' in chemsymbol:
            chemsymbol, sc = chemsymbol.split('.')
            assert sc == 'sc'
        self.symbol = chemsymbol
        self.type = hghdata.symbol
        self.name = 'LDA'
        self.initialize_setup_data()

    def initialize_setup_data(self):
        hghdata = self.hghdata
        beta = 0.1
        N = 450
        rgd = AERadialGridDescriptor(beta / N, 1.0 / N, N,
                                     default_spline_points=100)
        # rgd = EquidistantRadialGridDescriptor(0.001, 10000)
        self.rgd = rgd

        self.Z = hghdata.Z
        self.Nc = hghdata.Z - hghdata.Nv
        self.Nv = hghdata.Nv
        self.rcgauss = np.sqrt(2.0) * hghdata.rloc

        threshold = 1e-8
        if len(hghdata.c_n) > 0:
            vloc_g = create_local_shortrange_potential(rgd.r_g, hghdata.rloc,
                                                       hghdata.c_n)
            gcutvbar, rcutvbar = self.find_cutoff(rgd.r_g, rgd.dr_g, vloc_g,
                                                  threshold)
            self.vbar_g = np.sqrt(4.0 * np.pi) * vloc_g[:gcutvbar]
        else:
            rcutvbar = 0.5
            gcutvbar = rgd.ceil(rcutvbar)
            self.vbar_g = np.zeros(gcutvbar)

        nj = sum([v.nn for v in hghdata.v_l])
        if nj == 0:
            nj = 1  # Code assumes nj > 0 elsewhere, we fill out with zeroes

        if not hghdata.v_l:
            # No projectors.  But the remaining code assumes that everything
            # has projectors!  We'll just add the zero function then
            hghdata.v_l = [VNonLocal(0, 0.01, [0.])]

        n_j = []
        l_j = []

        # j ordering is significant, must be nl rather than ln
        for n, l in self.hghdata.nl_iter():
            n_j.append(n + 1)  # Note: actual n must be positive!
            l_j.append(l)
        assert nj == len(n_j)
        self.nj = nj
        self.l_j = l_j
        self.l_orb_J = l_j
        self.n_j = n_j

        self.rcut_j = []
        self.pt_jg = []

        for n, l in zip(n_j, l_j):
            # Note: even pseudopotentials without projectors will get one
            # projector, but the coefficients h_ij should be zero so it
            # doesn't matter
            pt_g = create_hgh_projector(rgd.r_g, l, n, hghdata.v_l[l].r0)
            norm = np.sqrt(np.dot(rgd.dr_g, pt_g**2 * rgd.r_g**2))
            assert np.abs(1 - norm) < 1e-5, str(1 - norm)
            gcut, rcut = self.find_cutoff(rgd.r_g, rgd.dr_g, pt_g, threshold)
            if rcut < 0.5:
                rcut = 0.5
                gcut = rgd.ceil(rcut)
            pt_g = pt_g[:gcut].copy()
            rcut = max(rcut, 0.5)
            self.rcut_j.append(rcut)
            self.pt_jg.append(pt_g)

        # This is the correct magnitude of the otherwise normalized
        # compensation charge
        self.Delta0 = -self.Nv / np.sqrt(4.0 * np.pi)

        f_ln = self.hghdata.get_occupation_numbers()
        f_j = [0] * nj
        for j, (n, l) in enumerate(self.hghdata.nl_iter()):
            try:
                f_j[j] = f_ln[l][n]
            except IndexError:
                pass
        self.f_ln = f_ln
        self.f_j = f_j

        r_g, lcomp, ghat = self.get_compensation_charge_functions()
        assert lcomp == [0] and len(ghat) == 1
        renormalized_ghat = self.Nv / (4.0 * np.pi) * ghat[0]
        self.Eh_compcharge = get_radial_hartree_energy(r_g, renormalized_ghat)

    def find_cutoff(self, r_g, dr_g, f_g, sqrtailnorm=1e-5):
        g = len(r_g)
        acc_sqrnorm = 0.0
        while acc_sqrnorm <= sqrtailnorm:
            g -= 1
            acc_sqrnorm += (r_g[g] * f_g[g])**2.0 * dr_g[g]
            if r_g[g] < 0.5:  # XXX
                return g, r_g[g]
        return g, r_g[g]

    def expand_hamiltonian_matrix(self):
        """Construct K_p from individual h_nn for each l."""
        ni = sum([2 * l + 1 for l in self.l_j])

        H_ii = np.zeros((ni, ni))

        # The H_ii used in gpaw is much larger and more general than the one
        # required for HGH pseudopotentials.  This means a lot of the elements
        # must be assigned the same value.  Not a performance issue though,
        # since these are small matrices
        M1start = 0
        for n1, l1 in self.hghdata.nl_iter():
            M1end = M1start + 2 * l1 + 1
            M2start = 0
            v = self.hghdata.v_l[l1]
            for n2, l2 in self.hghdata.nl_iter():
                M2end = M2start + 2 * l2 + 1
                if l1 == l2:
                    h_nn = v.expand_hamiltonian_diagonal()
                    H_mm = np.identity(M2end - M2start) * h_nn[n1, n2]
                    H_ii[M1start:M1end, M2start:M2end] += H_mm
                M2start = M2end
            M1start = M1end
        K_p = pack2(H_ii)
        return K_p

    def __str__(self):
        return "HGHSetupData('%s')" % self.type

    def __repr__(self):
        return self.__str__()

    def print_info(self, text, _setup):
        self.hghdata.print_info(text)

    def plot(self):
        """Plot localized functions of HGH setup."""
        import matplotlib.pyplot as plt
        rgd = self.rgd

        plt.subplot(211)  # vbar, compensation charge
        gcutvbar = len(self.vbar_g)
        plt.plot(rgd.r_g[:gcutvbar], self.vbar_g, 'r', label='vloc',
                 linewidth=3)
        rcc, gcc = self.get_compensation_charge_functions()
        gcc = gcc[0]

        plt.plot(rcc, gcc * self.Delta0, 'b--',
                 label='Comp charge [arb. unit]',
                 linewidth=3)
        plt.legend(loc='best')

        plt.subplot(212)  # projectors
        for j, (n, l, pt_g) in enumerate(zip(self.n_j, self.l_j, self.pt_jg)):
            label = 'n=%d, l=%d' % (n, l)
            plt.ylabel('$p_n^l(r)$')
            ng = len(pt_g)
            r_g = rgd.r_g[:ng]
            plt.plot(r_g, pt_g, label=label)
        plt.legend()

    def create_basis_functions(self):
        from gpaw.pseudopotential import generate_basis_functions
        return generate_basis_functions(self)

    def get_compensation_charge_functions(self):
        alpha = self.rcgauss**-2

        rcutgauss = self.rcgauss * 5.0
        # smaller values break charge conservation

        r = np.linspace(0.0, rcutgauss, 100)
        g = alpha**1.5 * np.exp(-alpha * r**2) * 4.0 / np.sqrt(np.pi)
        g[-1] = 0.0
        return r, [0], [g]

    def get_projectors(self):
        # XXX equal-range projectors still required for some reason
        maxlen = max([len(pt_g) for pt_g in self.pt_jg])
        pt_j = []
        for l, pt1_g in zip(self.l_j, self.pt_jg):
            pt2_g = self.rgd.zeros()[:maxlen]
            pt2_g[:len(pt1_g)] = pt1_g
            pt_j.append(self.rgd.spline(pt2_g, self.rgd.r_g[maxlen - 1], l))
        return pt_j

    def get_local_potential(self):
        n = len(self.vbar_g)
        return self.rgd.spline(self.vbar_g, self.rgd.r_g[n - 1], l=0)

    def build(self, xcfunc, lmax, basis, filter=None):
        if basis is None:
            basis = self.create_basis_functions()
        setup = PseudoPotential(self, basis)
        setup.fingerprint = hashlib.md5(str(self.hghdata).encode()).hexdigest()
        return setup


def create_local_shortrange_potential(r_g, rloc, c_n):
    rr_g = r_g / rloc  # "Relative r"
    rr2_g = rr_g**2
    rr4_g = rr2_g**2
    rr6_g = rr4_g * rr2_g

    gaussianpart = np.exp(-.5 * rr2_g)
    polypart = np.zeros(r_g.shape)
    for c, rrn_g in zip(c_n, [1, rr2_g, rr4_g, rr6_g]):
        polypart += c * rrn_g

    vloc_g = gaussianpart * polypart
    return vloc_g


def create_hgh_projector(r_g, l, n, r0):
    poly_g = r_g**(l + 2 * (n - 1))
    gauss_g = np.exp(-.5 * r_g**2 / r0**2)
    A = r0**(l + (4 * n - 1) / 2.0)
    assert (4 * n - 1) % 2 == 1
    B = half_integer_gamma[l + (4 * n - 1) // 2]**.5
    pt_g = 2.**.5 / A / B * poly_g * gauss_g
    return pt_g


# Coefficients determining off-diagonal elements of h_nn for l = 0...2
# given the diagonal elements
hcoefs_l = [
    [-.5 * (3. / 5.)**.5, .5 * (5. / 21.)**.5, -.5 * (100. / 63.)**.5],
    [-.5 * (5. / 7.)**.5, 1. / 6. * (35. / 11.)**.5, -1. / 6. * 14. / 11.**.5],
    [-.5 * (7. / 9.)**.5, .5 * (63. / 143)**.5, -.5 * 18. / 143.**.5]]


class VNonLocal:
    """Wrapper class for one nonlocal term of an HGH potential."""
    def __init__(self, l, r0, h_n):
        self.l = l
        self.r0 = r0
        h_n = np.array(h_n)
        nn = len(h_n)
        self.nn = nn
        self.h_n = h_n

    def expand_hamiltonian_diagonal(self):
        """Construct full atomic Hamiltonian from diagonal elements."""
        nn = self.nn
        h_n = self.h_n
        h_nn = np.zeros((nn, nn))
        for n, h in enumerate(h_n):
            h_nn[n, n] = h
        if self.l > 2:
            # print 'Warning: no diagonal elements for l=%d' % l
            # Some elements have projectors corresponding to l=3, but
            # the HGH article only specifies how to calculate the
            # diagonal elements of the atomic hamiltonian for l = 0, 1, 2 !
            return
        coefs = hcoefs_l[self.l]
        if nn > 2:
            h_nn[0, 2] = h_nn[2, 0] = coefs[1] * h_n[2]
            h_nn[1, 2] = h_nn[2, 1] = coefs[2] * h_n[2]
        if nn > 1:
            h_nn[0, 1] = h_nn[1, 0] = coefs[0] * h_n[1]
        return h_nn

    def copy(self):
        return VNonLocal(self.l, self.r0, self.h_n.copy())

    def serialize(self):  # no spin-orbit part
        return ' '.join(['    ', '%-10s' % self.r0] +
                        ['%10f' % h for h in self.h_n])


class HGHParameterSet:
    """Wrapper class for HGH-specific data corresponding to one element."""
    def __init__(self, symbol, Z, Nv, rloc, c_n, v_l):
        self.symbol = symbol  # Identifier, e.g. 'Na', 'Na.sc', ...
        self.Z = Z  # Actual atomic number
        self.Nv = Nv  # Valence electron count
        self.rloc = rloc  # Characteristic radius of local part
        self.c_n = np.array(c_n)  # Polynomial coefficients for local part
        self.v_l = list(v_l)  # Non-local parts

        Z, nlfe_j = configurations[self.symbol.split('.')[0]]
        self.configuration = nlfe_j

    def __str__(self):
        strings = ['HGH setup for %s\n' % self.symbol,
                   '    Valence Z=%d, rloc=%.05f\n' % (self.Nv, self.rloc)]

        if len(self.c_n) > 0:
            coef_string = ', '.join(['%.05f' % c for c in self.c_n])
        else:
            coef_string = 'zeros'
        strings.append('    Local part coeffs: %s\n' % coef_string)
        strings.append('    Projectors:\n')
        if not self.v_l:
            strings.append('        None\n')
        for v in self.v_l:
            strings.append('        l=%d, rc=%.05f\n' % (v.l, v.r0))
        strings.append('    Diagonal coefficients of nonlocal parts:')
        if not self.v_l:
            strings.append('\n        None\n')
        for v in self.v_l:
            strings.append('\n')
            strings.append('        l=%d: ' % v.l +
                           ', '.join(['%8.05f' % h for h in v.h_n]))
        return ''.join(strings)

    def copy(self):
        other = HGHParameterSet(self.symbol, self.Z, self.Nv, self.rloc,
                                self.c_n, self.v_l)
        return other

    def print_info(self, txt):
        txt(str(self))
        txt()

    def nl_iter(self):
        for n in range(4):
            for l, v in enumerate(self.v_l):
                if n < v.nn:
                    yield n, l

    def get_occupation_numbers(self):
        nlfe_j = list(self.configuration)
        nlfe_j.reverse()
        f_ln = [[], [], []]  # [[s], [p], [d]]
        # f states will be ignored as the atomic Hamiltonians
        # of those are, carelessly, not defined in the article.
        lmax = len(self.v_l) - 1
        Nv = 0
        # Right.  We need to find the occupation numbers of each state and
        # put them into a nice list of lists f_ln.
        #
        # We loop over states starting with the least bound one
        # (i.e. reversed nlfe_j), adding the occupation numbers of each state
        # as appropriate.  Once we have the right number of electrons, we
        # end the loop.
        #
        # Some states in the standard configuration might
        # be f-type; these should be skipped (unless the HGH setup actually
        # has a valence f-state; however as noted above, some of the
        # parameters are undefined in that case so are ignored anyway).  More
        # generally if for some state l > lmax,
        # we can skip that state.
        for n, l, f, e in nlfe_j:
            if l > lmax:
                continue
            Nv += f
            f_n = f_ln[l]
            assert f_n == [] or self.symbol.endswith('.sc')
            f_n.append(f)
            if Nv >= self.Nv:
                break
        assert Nv == self.Nv
        return f_ln

    def zeropad(self):
        """Return a new HGHParameterSet with all arrays zero padded so they
        have the same (max) length for all such HGH setups.  Makes
        plotting multiple HGH setups easier because they have compatible
        arrays."""
        c_n = np.zeros(4)
        for n, c in enumerate(self.c_n):
            c_n[n] = c
        v_l = []
        for l, v in enumerate(self.v_l):
            h_n = np.zeros(3)
            h_n[:len(v.h_n)] = list(v.h_n)
            v2 = VNonLocal(l, v.r0, h_n)
            v_l.append(v2)
        for l in range(len(self.v_l), 3):
            v_l.append(VNonLocal(l, 0.5, np.zeros(3)))
        copy = HGHParameterSet(self.symbol, self.Z, self.Nv, self.rloc, c_n,
                               v_l)
        return copy

    def serialize(self):
        string1 = '%-5s %-12s %10s ' % (self.symbol, self.Z, self.rloc)
        string2 = ' '.join(['%.10s' % c for c in self.c_n])
        nonlocal_strings = [v.serialize() for v in self.v_l]
        return '\n'.join([string1 + string2] + nonlocal_strings)


def parse_local_part(string):
    """Create HGHParameterSet object with local part initialized."""
    tokens = iter(string.split())
    symbol = next(tokens)
    actual_chemical_symbol = symbol.split('.')[0]
    Z = atomic_numbers[actual_chemical_symbol]
    Nv = int(next(tokens))
    rloc = float(next(tokens))
    c_n = [float(token) for token in tokens]
    return symbol, Z, Nv, rloc, c_n


class HGHBogusNumbersError(ValueError):
    """Error which is raised when the HGH parameters contain f-type
    or higher projectors.  The HGH article only defines atomic Hamiltonian
    matrices up to l=2, so these are meaningless."""
    pass


def parse_hgh_setup(lines):
    """Initialize HGHParameterSet object from text representation."""
    lines = iter(lines)
    symbol, Z, Nv, rloc, c_n = parse_local_part(next(lines))

    def pair_up_nonlocal_lines(lines):
        try:
            yield (next(lines), '')
            while True:
                yield (next(lines), next(lines))
        except StopIteration:
            return

    v_l = []
    for l, (non_local, spinorbit) in enumerate(pair_up_nonlocal_lines(lines)):
        # we discard the spinorbit 'k_n' data so far
        nltokens = non_local.split()
        r0 = float(nltokens[0])
        h_n = [float(token) for token in nltokens[1:]]

        # if h_n[-1] == 0.0: # Only spin-orbit contributes.  Discard.
        #     h_n.pop()
        # Actually the above causes trouble.  Probably it messes up state
        # ordering or something else that shouldn't have any effect.

        vnl = VNonLocal(l, r0, h_n)
        v_l.append(vnl)
        if l > 2:
            raise HGHBogusNumbersError

    hgh = HGHParameterSet(symbol, Z, Nv, rloc, c_n, v_l)
    return hgh


def str2hgh(string):
    return parse_hgh_setup(string.splitlines())


def hgh2str(hgh):
    return hgh.serialize()


def parse_setups(lines):
    """Read HGH data from file."""
    setups = {}
    entry_lines = [i for i in range(len(lines))
                   if lines[i][0].isalpha()]
    lines_by_element = [lines[entry_lines[i]:entry_lines[i + 1]]
                        for i in range(len(entry_lines) - 1)]
    lines_by_element.append(lines[entry_lines[-1]:])

    for elines in lines_by_element:
        try:
            hgh = parse_hgh_setup(elines)
        except HGHBogusNumbersError:
            continue
        assert hgh.symbol not in setups
        setups[hgh.symbol] = hgh
    return setups


def plot(symbol, extension=None):
    import matplotlib.pyplot as plt
    try:
        s = HGHSetupData(symbol)
    except IndexError:
        print('Nooooo')
        return
    s.plot()
    if extension is not None:
        plt.savefig('hgh.%s.%s' % (symbol, extension))


def plot_many(*symbols):
    import matplotlib.pyplot as plt
    if not symbols:
        symbols = setups.keys() + [key + '.sc' for key in sc_setups.keys()]
    for symbol in symbols:
        plt.figure(1)
        plot(symbol, extension='png')
        plt.clf()


def parse_default_setups():
    from gpaw.hgh_parameters import parameters
    lines = parameters.splitlines()
    setups0 = parse_setups(lines)
    for key, value in setups0.items():
        if key.endswith('.sc'):
            sym, sc = key.split('.')
            sc_setups[sym] = value
        else:
            setups[key] = value


parse_default_setups()



from __future__ import annotations

from math import pi
from typing import TYPE_CHECKING

import numpy as np
from ase.units import Ha

import gpaw.fftw as fftw
from gpaw import debug
from gpaw.core.arrays import DistributedArrays
from gpaw.core.domain import Domain
from gpaw.core.matrix import Matrix
from gpaw.core.pwacf import PWAtomCenteredFunctions
from gpaw.gpu import cupy as cp
from gpaw.mpi import MPIComm, serial_comm
from gpaw.new import prod, zips
from gpaw.new.c import (add_to_density, add_to_density_gpu, pw_insert,
                        pw_insert_gpu)
from gpaw.pw.descriptor import pad
from gpaw.typing import (Array1D, Array2D, Array3D, ArrayLike1D, ArrayLike2D,
                         Literal, Vector)

if TYPE_CHECKING:
    from gpaw.core import UGArray, UGDesc


class PWDesc(Domain):
    itemsize = 16

    def __init__(self,
                 *,
                 ecut: float,  # hartree
                 cell: ArrayLike1D | ArrayLike2D,  # bohr
                 kpt: Vector | None = None,  # in units of reciprocal cell
                 comm: MPIComm = serial_comm,
                 dtype=None):
        """Description of plane-wave basis.

        parameters
        ----------
        ecut:
            Cutoff energy for kinetic energy of plane waves (units: hartree).
        cell:
            Unit cell given as three floats (orthorhombic grid), six floats
            (three lengths and the angles in degrees) or a 3x3 matrix
            (units: bohr).
        comm:
            Communicator for distribution of plane-waves.
        kpt:
            K-point for Block-boundary conditions specified in units of the
            reciprocal cell.
        dtype:
            Data-type (float or complex).
        """
        self.ecut = ecut
        Domain.__init__(self, cell, (True, True, True), kpt, comm, dtype)

        G_plus_k_Gv, ekin_G, self.indices_cG = find_reciprocal_vectors(
            ecut, self.cell_cv, self.kpt_c, self.dtype)

        # Find distribution:
        S = comm.size
        ng = len(ekin_G)
        self.maxmysize = (ng + S - 1) // S
        ng1 = comm.rank * self.maxmysize
        ng2 = min(ng1 + self.maxmysize, ng)
        self.ng1 = ng1
        self.ng2 = ng2

        # Distribute things:
        self.ekin_G = ekin_G[ng1:ng2].copy()
        self.ekin_G.flags.writeable = False
        # self.myindices_cG = self.indices_cG[:, ng1:ng2]
        self.G_plus_k_Gv = G_plus_k_Gv[ng1:ng2].copy()

        self.shape = (ng,)
        self.myshape = (len(self.ekin_G),)

        # Convert from np.float64 to float to avoid fake cupy problem ...
        # XXX Fix cpupy!!!
        self.dv = float(abs(np.linalg.det(self.cell_cv)))

        self._indices_cache: dict[tuple[int, ...], Array1D] = {}

        self.qspiral_v = None

    def __repr__(self) -> str:
        m = self.myshape[0]
        n = self.shape[0]
        r = Domain.__repr__(self).replace(
            'Domain(',
            f'PWDesc(ecut={self.ecut} <coefs={m}/{n}>, ')
        if self.qspiral_v is None:
            return r
        q = self.cell_cv @ self.qspiral_v / (2 * pi)
        return f'{r[:-1]}, qsiral={q}'

    def _short_string(self, global_shape):
        return (f'plane wave coefficients: {global_shape[-1]}\n'
                f'cutoff: {self.ecut * Ha} eV\n')

    def global_shape(self) -> tuple[int, ...]:
        """Tuple with one element: number of plane waves."""
        return self.shape

    def reciprocal_vectors(self) -> Array2D:
        """Returns reciprocal lattice vectors, G + k, in xyz coordinates."""
        return self.G_plus_k_Gv

    def kinetic_energies(self) -> Array1D:
        """Kinetic energy of plane waves.

        :::

             _ _ 2
            |G+k| / 2

        """
        return self.ekin_G

    def empty(self,
              dims: int | tuple[int, ...] = (),
              comm: MPIComm = serial_comm,
              xp=None) -> PWArray:
        """Create new PlaneWaveExpanions object.

        parameters
        ----------
        dims:
            Extra dimensions.
        comm:
            Distribute dimensions along this communicator.
        """
        return PWArray(self, dims, comm, xp=xp)

    def from_data(self, data):
        return PWArray(self, data.shape[:-1], data=data)

    def new(self,
            *,
            ecut: float | None = None,
            kpt=None,
            dtype=None,
            comm: MPIComm | Literal['inherit'] | None = 'inherit'
            ) -> PWDesc:
        """Create new plane-wave expansion description."""
        comm = self.comm if comm == 'inherit' else comm or serial_comm
        return PWDesc(ecut=ecut or self.ecut,
                      cell=self.cell_cv,
                      kpt=self.kpt_c if kpt is None else kpt,
                      dtype=dtype or self.dtype,
                      comm=comm or serial_comm)

    def indices(self, shape: tuple[int, ...]) -> Array1D:
        """Return indices into FFT-grid."""
        Q_G = self._indices_cache.get(shape)
        if Q_G is None:
            Q_G = np.ravel_multi_index(self.indices_cG, shape,  # type: ignore
                                       mode='wrap').astype(np.int32)
            self._indices_cache[shape] = Q_G
        return Q_G

    def cut(self, array_Q: Array3D) -> Array1D:
        """Cut out G-vectors with (G+k)^2/2<E_kin."""
        return array_Q.ravel()[self.indices(array_Q.shape)]

    def paste(self, coef_G: Array1D, array_Q: Array3D) -> None:
        """Paste G-vectors with (G+k)^2/2<E_kin into 3-D FFT grid and
        zero-pad."""
        Q_G = self.indices(array_Q.shape)
        if debug:
            assert (Q_G[1:] > Q_G[:-1]).all()
            assert (Q_G >= 0).all()
            assert (Q_G < array_Q.size).all()
            assert coef_G.shape == Q_G.shape
            assert coef_G.flags.c_contiguous
            assert Q_G.flags.c_contiguous
            assert array_Q.flags.c_contiguous

        pw_insert(coef_G, Q_G, 1.0, array_Q)

    def map_indices(self, other: PWDesc) -> tuple[Array1D, list[Array1D]]:
        """Map from one (distributed) set of plane waves to smaller global set.

        Say we have 9 G-vector on two cores::

           5 3 4             . 3 4           0 . .
           2 0 1 -> rank=0:  2 0 1  rank=1:  . . .
           8 6 7             . . .           3 1 2

        and we want a mapping to these 5 G-vectors::

             3
           2 0 1
             4

        On rank=0: the return values are::

           [0, 1, 2, 3], [[0, 1, 2, 3], [4]]

        and for rank=1::

           [1], [[0, 1, 2, 3], [4]]
        """
        size_c = tuple(self.indices_cG.ptp(axis=1) + 1)  # type: ignore
        Q_G = self.indices(size_c)
        G_Q = np.empty(prod(size_c), int)
        G_Q[Q_G] = np.arange(len(Q_G))
        G_g = G_Q[other.indices(size_c)]
        ng1 = 0
        g_r = []
        for rank in range(self.comm.size):
            ng2 = min(ng1 + self.maxmysize, self.shape[0])
            myg = (ng1 <= G_g) & (G_g < ng2)
            g_r.append(np.nonzero(myg)[0])
            if rank == self.comm.rank:
                my_G_g = G_g[myg] - ng1
            ng1 = ng2
        return my_G_g, g_r

    def atom_centered_functions(self,
                                functions,
                                positions,
                                *,
                                atomdist=None,
                                integral=None,
                                cut=False,
                                xp=None):
        """Create PlaneWaveAtomCenteredFunctions object."""
        if self.qspiral_v is None:
            return PWAtomCenteredFunctions(functions, positions, self,
                                           atomdist=atomdist,
                                           xp=xp)

        from gpaw.new.spinspiral import SpiralPWACF
        return SpiralPWACF(functions, positions, self,
                           atomdist=atomdist,
                           qspiral_v=self.qspiral_v)


class PWArray(DistributedArrays[PWDesc]):
    def __init__(self,
                 pw: PWDesc,
                 dims: int | tuple[int, ...] = (),
                 comm: MPIComm = serial_comm,
                 data: np.ndarray | None = None,
                 xp=None):
        """Object for storing function(s) as a plane-wave expansions.

        parameters
        ----------
        pw:
            Description of plane-waves.
        dims:
            Extra dimensions.
        comm:
            Distribute plane-waves along this communicator.
        data:
            Data array for storage.
        """
        DistributedArrays. __init__(self, dims, pw.myshape,
                                    comm, pw.comm,
                                    data, pw.dv, complex, xp)
        self.desc = pw
        self._matrix: Matrix | None

    def __repr__(self):
        txt = f'PWArray(pw={self.desc}, dims={self.dims}'
        if self.comm.size > 1:
            txt += f', comm={self.comm.rank}/{self.comm.size}'
        if self.xp is not np:
            txt += ', xp=cp'
        return txt + ')'

    def __getitem__(self, index: int | slice) -> PWArray:
        data = self.data[index]
        return PWArray(self.desc, data.shape[:-1], data=data)

    def __iter__(self):
        for data in self.data:
            yield PWArray(self.desc,
                          data.shape[:-len(self.desc.shape)],
                          data=data)

    def new(self,
            data=None) -> PWArray:
        """Create new PWArray object of same kind.

        Parameters
        ----------
        data:
            Array to use for storage.
        """
        if data is None:
            data = self.xp.empty_like(self.data)
        else:
            # Number of plane-waves depends on the k-point.  We therfore
            # allow for data to be bigger than needed:
            data = data.ravel()[:self.data.size].reshape(self.data.shape)
        return PWArray(self.desc,
                       self.dims,
                       self.comm,
                       data)

    def copy(self):
        """Create a copy (surprise!)."""
        a = self.new()
        a.data[:] = self.data
        return a

    def _arrays(self):
        shape = self.data.shape
        return self.data.reshape((prod(shape[:-1]), shape[-1]))

    @property
    def matrix(self) -> Matrix:
        """Matrix view of data."""
        if self._matrix is not None:
            return self._matrix

        shape = (self.dims[0], prod(self.dims[1:]) * self.myshape[0])
        myshape = (self.mydims[0], prod(self.mydims[1:]) * self.myshape[0])
        dist = (self.comm, -1, 1)
        data = self.data.reshape(myshape)

        if self.desc.dtype == float:
            data = data.view(float)
            shape = (shape[0], shape[1] * 2)

        self._matrix = Matrix(*shape, data=data, dist=dist)
        return self._matrix

    def ifft(self, *, plan=None, grid=None, out=None, periodic=False):
        """Do inverse FFT(s) to uniform grid(s).

        Parameters
        ----------
        plan:
            Plan for inverse FFT.
        grid:
            Target grid.
        out:
            Target UGArray object.
        """
        comm = self.desc.comm
        xp = self.xp
        if out is None:
            out = grid.empty(self.dims, xp=xp)
        assert self.desc.dtype == out.desc.dtype, (self.desc, out.desc)
        assert out.desc.pbc_c.all()
        assert comm.size == out.desc.comm.size

        plan = plan or out.desc.fft_plans(xp=xp)
        this = self.gather()
        if this is not None:
            for coef_G, out1 in zips(this._arrays(), out.flat()):
                plan.ifft_sphere(coef_G, self.desc, out1)
        else:
            for out1 in out.flat():
                plan.ifft_sphere(None, self.desc, out1)

        if not periodic:
            out.multiply_by_eikr()

        return out

    def interpolate(self,
                    plan1: fftw.FFTPlans | None = None,
                    plan2: fftw.FFTPlans | None = None,
                    grid: UGDesc | None = None,
                    out: UGArray | None = None) -> UGArray:
        assert plan1 is None
        return self.ifft(plan=plan2, grid=grid, out=out)

    def gather(self, out=None, broadcast=False):
        """Gather coefficients on master."""
        comm = self.desc.comm

        if comm.size == 1:
            if out is None:
                return self
            out.data[:] = self.data
            return out

        if out is None:
            if comm.rank == 0 or broadcast:
                pw = self.desc.new(comm=serial_comm)
                out = pw.empty(self.dims, comm=self.comm, xp=self.xp)
            else:
                out = Empty(self.mydims)

        if comm.rank == 0:
            data = self.xp.empty(self.desc.maxmysize * comm.size, complex)
        else:
            data = None

        for input, output in zips(self._arrays(), out._arrays()):
            mydata = pad(input, self.desc.maxmysize)
            comm.gather(mydata, 0, data)
            if comm.rank == 0:
                output[:] = data[:len(output)]

        if broadcast:
            comm.broadcast(out.data, 0)

        return out if not isinstance(out, Empty) else None

    def gather_all(self, out: PWArray) -> None:
        """Gather coefficients from self[r] on rank r.

        On rank r, an array of all G-vector coefficients will be returned.
        These will be gathered from self[r] on all the cores.
        """
        assert len(self.dims) == 1
        pw = self.desc
        comm = pw.comm
        if comm.size == 1:
            out.data[:] = self.data[0]
            return

        N = self.dims[0]
        assert N <= comm.size

        ng = pw.shape[0]
        myng = pw.myshape[0]
        maxmyng = pw.maxmysize

        ssize_r, soffset_r, rsize_r, roffset_r = a2a_stuff(
            comm, N, ng, myng, maxmyng)

        comm.alltoallv(self.data, ssize_r, soffset_r,
                       out.data, rsize_r, roffset_r)

    def scatter_from(self, data: Array1D | PWArray | None = None) -> None:
        """Scatter data from rank-0 to all ranks."""
        if isinstance(data, PWArray):
            data = data.data
        comm = self.desc.comm
        if comm.size == 1:
            assert data is not None
            self.data[:] = self.xp.asarray(data)
            return

        assert self.dims == ()

        if comm.rank == 0:
            data = pad(data, comm.size * self.desc.maxmysize)
            comm.scatter(data, self.data, 0)
        else:
            buf = self.xp.empty(self.desc.maxmysize, complex)
            comm.scatter(None, buf, 0)
            self.data[:] = buf[:len(self.data)]

    def scatter_from_all(self, a_G: PWArray) -> None:
        """Scatter all coefficients from rank r to self on other cores."""
        assert len(self.dims) == 1
        pw = self.desc
        comm = pw.comm
        if comm.size == 1:
            self.data[:] = a_G.data
            return

        N = self.dims[0]
        assert N <= comm.size

        ng = pw.shape[0]
        myng = pw.myshape[0]
        maxmyng = pw.maxmysize

        rsize_r, roffset_r, ssize_r, soffset_r = a2a_stuff(
            comm, N, ng, myng, maxmyng)

        comm.alltoallv(a_G.data, ssize_r, soffset_r,
                       self.data, rsize_r, roffset_r)

    def integrate(self, other: PWArray | None = None) -> np.ndarray:
        """Integral of self or self time cc(other)."""
        dv = self.dv
        if other is not None:
            assert self.comm.size == 1
            assert self.desc.dtype == other.desc.dtype
            a = self._arrays()
            b = other._arrays()
            if self.desc.dtype == float:
                a = a.view(float)
                b = b.view(float)
                dv *= 2
            result = a @ b.T.conj()
            if self.desc.dtype == float and self.desc.comm.rank == 0:
                result -= 0.5 * a[:, :1] @ b[:, :1].T
            self.desc.comm.sum(result)
            result = result.reshape(self.dims + other.dims)
        else:
            if self.desc.comm.rank == 0:
                result = self.data[..., 0]
            else:
                result = self.xp.empty(self.mydims, complex)
            self.desc.comm.broadcast(result, 0)

        if self.desc.dtype == float:
            result = result.real
        if result.ndim == 0:
            result = result.item()  # convert to scalar
        return result * dv

    def _matrix_elements_correction(self,
                                    M1: Matrix,
                                    M2: Matrix,
                                    out: Matrix,
                                    symmetric: bool) -> None:
        if self.desc.dtype == float:
            out.data *= 2.0
            if self.desc.comm.rank == 0:
                correction = M1.data[:, :1] @ M2.data[:, :1].T
                if symmetric:
                    correction *= 0.5 * self.dv
                    out.data -= correction
                    out.data -= correction.T
                else:
                    correction *= self.dv
                    out.data -= correction

    def norm2(self, kind: str = 'normal') -> np.ndarray:
        r"""Calculate integral over cell.

        For kind='normal' we calculate:::

          /   _  2 _   --    2
          ||a(r)| dr = > |c | V,
          /            --  G
                        G

        where V is the volume of the unit cell.

        And for kind='kinetic':::

           1  --    2  2
          --- > |c |  G V,
           2  --  G
               G

        """
        a_xG = self._arrays().view(float)
        if kind == 'normal':
            result_x = self.xp.einsum('xG, xG -> x', a_xG, a_xG)
        elif kind == 'kinetic':

            a_xG = a_xG.reshape((len(a_xG), -1, 2))
            result_x = self.xp.einsum('xGi, xGi, G -> x',
                                      a_xG,
                                      a_xG,
                                      self.xp.asarray(self.desc.ekin_G))
        else:
            1 / 0
        if self.desc.dtype == float:
            result_x *= 2
            if self.desc.comm.rank == 0 and kind == 'normal':
                result_x -= a_xG[:, 0]**2
        self.desc.comm.sum(result_x)
        return result_x.reshape(self.mydims) * self.dv

    def abs_square(self,
                   weights: Array1D,
                   out: UGArray,
                   _slow: bool = False) -> None:
        """Add weighted absolute square of self to output array.

        With `a_n(G)` being self and `w_n` the weights:::

              _         _    --     -1    _   2
          out(r) <- out(r) + >  |FFT  [a (G)]| w
                             --         n       n
                             n

        """
        pw = self.desc
        domain_comm = pw.comm
        xp = self.xp
        a_nG = self

        if domain_comm.size == 1:
            if not _slow and xp is cp and pw.dtype == complex:
                return abs_square_gpu(a_nG, weights, out)

            a_R = out.desc.new(dtype=pw.dtype).empty(xp=xp)
            for weight, a_G in zips(weights, a_nG):
                if weight == 0.0:
                    continue
                a_G.ifft(out=a_R)
                if xp is np:
                    add_to_density(weight, a_R.data, out.data)
                else:
                    out.data += float(weight) * xp.abs(a_R.data)**2
            return

        # Undistributed work arrays:
        a1_R = out.desc.new(comm=None, dtype=pw.dtype).empty(xp=xp)
        a1_G = pw.new(comm=None).empty(xp=xp)
        b1_R = out.desc.new(comm=None).zeros(xp=xp)

        (N,) = self.mydims
        for n1 in range(0, N, domain_comm.size):
            n2 = min(n1 + domain_comm.size, N)
            a_nG[n1:n2].gather_all(a1_G)
            n = n1 + domain_comm.rank
            if n >= N:
                continue
            weight = weights[n]
            if weight == 0.0:
                continue
            a1_G.ifft(out=a1_R)
            if xp is np:
                add_to_density(weight, a1_R.data, b1_R.data)
            else:
                b1_R.data += float(weight) * xp.abs(a1_R.data)**2

        domain_comm.sum(b1_R.data)
        b_R = out.new()
        b_R.scatter_from(b1_R)
        out.data += b_R.data

    def to_pbc_grid(self):
        return self

    def randomize(self, seed: int | None = None) -> None:
        """Insert random numbers between -0.5 and 0.5 into data."""
        if seed is None:
            seed = self.comm.rank + self.desc.comm.rank * self.comm.size
        rng = self.xp.random.default_rng(seed)
        a = self.data.view(float)
        rng.random(a.shape, out=a)
        a -= 0.5
        if self.desc.dtype == float and self.desc.comm.rank == 0:
            a[..., 1] = 0.0

    def moment(self):
        pw = self.desc
        # Masks:
        m0_G, m1_G, m2_G = [i_G == 0 for i_G in pw.indices_cG]
        a_G = self.gather()
        if a_G is not None:
            b_G = a_G.data.imag
            b_cs = [b_G[m1_G & m2_G],
                    b_G[m0_G & m2_G],
                    b_G[m0_G & m1_G]]
            d_c = [b_s[1:] @ (1.0 / np.arange(1, len(b_s)))
                   for b_s in b_cs]
            m_v = np.dot(d_c, pw.cell_cv) / pi * pw.dv
        else:
            m_v = np.empty(3)
        pw.comm.broadcast(m_v, 0)
        return m_v

    def morph(self, pw: PWDesc) -> PWArray:
        pw0 = self.desc
        out_xG = pw.zeros(self.dims,
                          comm=self.comm,
                          xp=self.xp)
        assert isinstance(out_xG, PWArray)  # MYPY!!!!

        d = {}
        for G, i_c in enumerate(pw.indices_cG.T):
            d[tuple(i_c)] = G
        G_G0 = []
        G0_G = []
        for G0, i_c in enumerate(pw0.indices_cG.T):
            G = d.get(tuple(i_c), -1)
            if G != -1:
                G_G0.append(G)
                G0_G.append(G0)

        out_xG.data[..., G_G0] = self.data[..., G0_G]
        return out_xG

    def add_ked(self,
                occ_n: Array1D,
                taut_R: UGArray) -> None:
        psit_nG = self
        pw = psit_nG.desc
        domain_comm = pw.comm

        # Undistributed work arrays:
        dpsit1_R = taut_R.desc.new(comm=None, dtype=pw.dtype).empty()
        pw1 = pw.new(comm=None)
        psit1_G = pw1.empty()
        iGpsit1_G = pw1.empty()
        taut1_R = taut_R.desc.new(comm=None).zeros()
        Gplusk1_Gv = pw1.reciprocal_vectors()

        (N,) = psit_nG.mydims
        for n1 in range(0, N, domain_comm.size):
            n2 = min(n1 + domain_comm.size, N)
            psit_nG[n1:n2].gather_all(psit1_G)
            n = n1 + domain_comm.rank
            if n >= N:
                continue
            f = occ_n[n]
            if f == 0.0:
                continue
            for v in range(3):
                iGpsit1_G.data[:] = psit1_G.data
                iGpsit1_G.data *= 1j * Gplusk1_Gv[:, v]
                iGpsit1_G.ifft(out=dpsit1_R)
                add_to_density(0.5 * f, dpsit1_R.data, taut1_R.data)
        domain_comm.sum(taut1_R.data)
        tmp_R = taut_R.new()
        tmp_R.scatter_from(taut1_R)
        taut_R.data += tmp_R.data


def a2a_stuff(comm, N, ng, myng, maxmyng):
    """Create arrays for MPI alltoallv call."""
    ssize_r = np.zeros(comm.size, int)
    ssize_r[:N] = myng
    soffset_r = np.arange(comm.size) * myng
    soffset_r[N:] = 0
    roffset_r = (np.arange(comm.size) * maxmyng).clip(max=ng)
    rsize_r = np.zeros(comm.size, int)
    if comm.rank < N:
        rsize_r[:-1] = roffset_r[1:] - roffset_r[:-1]
        rsize_r[-1] = ng - roffset_r[-1]
    return ssize_r, soffset_r, rsize_r, roffset_r


class Empty:
    def __init__(self, dims):
        self.dims = dims

    def _arrays(self):
        for _ in range(prod(self.dims)):
            yield


def find_reciprocal_vectors(ecut: float,
                            cell: Array2D,
                            kpt=np.zeros(3),
                            dtype=complex) -> tuple[Array2D,
                                                    Array1D,
                                                    Array2D]:
    """Find reciprocal lattice vectors inside sphere.

    >>> cell = np.eye(3)
    >>> ecut = 0.5 * (2 * pi)**2
    >>> G, e, i = find_reciprocal_vectors(ecut, cell)
    >>> G
    array([[ 0.        ,  0.        ,  0.        ],
           [ 0.        ,  0.        ,  6.28318531],
           [ 0.        ,  0.        , -6.28318531],
           [ 0.        ,  6.28318531,  0.        ],
           [ 0.        , -6.28318531,  0.        ],
           [ 6.28318531,  0.        ,  0.        ],
           [-6.28318531,  0.        ,  0.        ]])
    >>> e
    array([ 0.       , 19.7392088, 19.7392088, 19.7392088, 19.7392088,
           19.7392088, 19.7392088])
    >>> i
    array([[ 0,  0,  0,  0,  0,  1, -1],
           [ 0,  0,  0,  1, -1,  0,  0],
           [ 0,  1, -1,  0,  0,  0,  0]])
    """
    Gcut = (2 * ecut)**0.5
    n = Gcut * (cell**2).sum(axis=1)**0.5 / (2 * pi) + abs(kpt)
    size = 2 * n.astype(int) + 4

    if dtype == float:
        size[2] = size[2] // 2 + 1
        i_Qc = np.indices(size).transpose((1, 2, 3, 0))
        i_Qc[..., :2] += size[:2] // 2
        i_Qc[..., :2] %= size[:2]
        i_Qc[..., :2] -= size[:2] // 2
    else:
        i_Qc = np.indices(size).transpose((1, 2, 3, 0))  # type: ignore
        half = [s // 2 for s in size]
        i_Qc += half
        i_Qc %= size
        i_Qc -= half

    # Calculate reciprocal lattice vectors:
    B_cv = 2.0 * pi * np.linalg.inv(cell).T
    # i_Qc.shape = (-1, 3)
    G_plus_k_Qv = (i_Qc + kpt) @ B_cv

    ekin = 0.5 * (G_plus_k_Qv**2).sum(axis=3)
    mask = ekin <= ecut

    assert not mask[size[0] // 2].any()
    assert not mask[:, size[1] // 2].any()
    if dtype == complex:
        assert not mask[:, :, size[2] // 2].any()
    else:
        assert not mask[:, :, -1].any()

    if dtype == float:
        mask &= ((i_Qc[..., 2] > 0) |
                 (i_Qc[..., 1] > 0) |
                 ((i_Qc[..., 0] >= 0) & (i_Qc[..., 1] == 0)))

    indices = i_Qc[mask]
    ekin = ekin[mask]
    G_plus_k = G_plus_k_Qv[mask]

    return G_plus_k, ekin, indices.T


def abs_square_gpu(psit_nG, weight_n, nt_R):
    from gpaw.gpu import cupyx
    pw = psit_nG.desc
    plan = nt_R.desc.fft_plans(xp=cp, dtype=complex)
    Q_G = cp.asarray(plan.indices(pw))
    weight_n = cp.asarray(weight_n)
    N = len(weight_n)
    shape = tuple(nt_R.desc.size_c)
    B = 32
    psit_bR = None
    for b1 in range(0, N, B):
        b2 = min(b1 + B, N)
        nb = b2 - b1
        if psit_bR is None:
            psit_bR = cp.empty((nb,) + shape, complex)
        elif nb < B:
            psit_bR = psit_bR[:nb]
        psit_bR[:] = 0.0
        pw_insert_gpu(psit_nG.data[b1:b2],
                      Q_G,
                      1.0,
                      psit_bR.reshape((nb, -1)))
        psit_bR[:] = cupyx.scipy.fft.ifftn(
            psit_bR,
            shape,
            norm='forward',
            overwrite_x=True)
        add_to_density_gpu(weight_n[b1:b2], psit_bR, nt_R.data)



from ase import Atoms
from ase.calculators.test import numeric_force
from gpaw import GPAW, FermiDirac, PoissonSolver
from gpaw.test import equal
from gpaw.xc.tools import vxc


def test_generic_8Si():
    a = 5.404
    bulk = Atoms(symbols='Si8',
                 positions=[(0, 0, 0.1 / a),
                            (0, 0.5, 0.5),
                            (0.5, 0, 0.5),
                            (0.5, 0.5, 0),
                            (0.25, 0.25, 0.25),
                            (0.25, 0.75, 0.75),
                            (0.75, 0.25, 0.75),
                            (0.75, 0.75, 0.25)],
                 pbc=True)
    bulk.set_cell((a, a, a), scale_atoms=True)
    n = 20
    calc = GPAW(mode='fd',
                gpts=(n, n, n),
                nbands='150%',
                occupations=FermiDirac(width=0.01),
                poissonsolver=PoissonSolver('fd', nn='M', relax='J'),
                kpts=(2, 2, 2),
                convergence={'energy': 1e-7}
                )
    bulk.calc = calc
    f1 = bulk.get_forces()[0, 2]
    e1 = bulk.get_potential_energy()
    v_xc = vxc(calc.gs_adapter())
    print(v_xc)
    niter1 = calc.get_number_of_iterations()

    f2 = numeric_force(bulk, 0, 2)
    print((f1, f2, f1 - f2))
    equal(f1, f2, 0.005)

    # Volume per atom:
    vol = a**3 / 8
    de = calc.get_electrostatic_corrections() / vol
    print(de)
    assert abs(de[0] - -2.190) < 0.001

    print((e1, f1, niter1))
    energy_tolerance = 0.0025
    force_tolerance = 0.01
    equal(e1, -46.6628, energy_tolerance)  # svnversion 5252
    equal(f1, -1.38242356123, force_tolerance)  # svnversion 5252



"""Check URL's in Python files."""
import re
import sys
from pathlib import Path
from urllib.error import HTTPError, URLError
from urllib.request import urlopen, Request

OK = {'https://doi.org/%s',
      'https://arxiv.org/abs/%s',
      'https://gitlab.com/gpaw/gpaw/-/merge_requests/%s',
      'https://gitlab.com/gpaw/gpaw/-/issues/%s',
      'https://xkcd.com/%s',
      'https://gitlab.com/ase/ase.git@master',
      'https://gitlab.com/{name}/{name}.git',
      'https://cmrdb.fysik.dtu.dk/c2db',
      'https://wiki.fysik.dtu.dk/gpaw-files/gpaw-setups-*.tar.gz',
      'https://wiki.fysik.dtu.dk/gpaw-files',
      'https://wiki.fysik.dtu.dk/gpaw-files/',
      'https://wiki.fysik.dtu.dk/gpaw-files/things/',
      'https://wiki.fysik.dtu.dk/gpaw/devel'}

USERAGENT = 'Mozilla/5.0 (X11; U; Linux i686) Gecko/20071127 Firefox/2.0.0.11'


def check(root: Path) -> int:
    """Chech URL's in Python files inside root."""
    errors = 0
    for path in root.glob('**/*.py'):
        for n, line in enumerate(path.read_text().splitlines()):
            for url in re.findall(r'https?://\S+', line):
                url = url.rstrip(""",.'"}):""")
                if url not in OK and 'html/_downloads' not in str(path):
                    if '(' in url and ')' not in url:
                        url += ')'
                    if not check1(path, n, url):
                        errors += 1
    return errors


def check1(path: Path, n: int, url: str) -> bool:
    try:
        req = Request(url, headers={'User-Agent': USERAGENT})
        urlopen(req)
    except (HTTPError, URLError, ConnectionResetError) as e:
        print(f'{path}:{n + 1}')
        print(url)
        print(e)
        print()
        return False
    except Exception:
        print(f'{path}:{n + 1}')
        print(url)
        raise
    return True


def test():
    errors = sum(check(Path(f)) for f in ['gpaw', 'doc'])
    assert errors < 10


if __name__ == '__main__':
    for arg in sys.argv[1:]:
        root = Path(arg)
        check(root)



"""Provide mechanism to broadcast imports from master to other processes.

This reduces file system strain.

Use:

  with broadcast_imports():
      <execute import statements>

This temporarily overrides the Python import mechanism so that

  1) master executes and caches import metadata and code
  2) import metadata and code are broadcast to all processes
  3) other processes execute the import statements from memory

Warning: Do not perform any parallel operations while broadcast imports
are enabled.  Non-master processes assume that they will receive module
data and will crash or deadlock if master sends anything else.
"""


import os
import sys
import marshal
from importlib.machinery import PathFinder, ModuleSpec

import _gpaw


if hasattr(_gpaw, 'Communicator'):
    if '_gpaw' not in sys.builtin_module_names:
        libmpi = os.environ.get('GPAW_MPI', 'libmpi.so')
        import ctypes
        try:
            ctypes.CDLL(libmpi, ctypes.RTLD_GLOBAL)
        except OSError:
            pass
    world = _gpaw.Communicator()
else:
    world = None  # type: ignore


def marshal_broadcast(obj):
    if world.rank == 0:
        buf = marshal.dumps(obj)
    else:
        assert obj is None
        buf = None

    buf = _gpaw.globally_broadcast_bytes(buf)
    try:
        return marshal.loads(buf)
    except ValueError as err:
        msg = ('Parallel import failure -- probably received garbage.  '
               'Error was: {}.  This may happen if parallel operations are '
               'performed while parallel imports are enabled.'.format(err))
        raise ImportError(msg)


class BroadcastLoader:
    def __init__(self, spec, module_cache):
        self.module_cache = module_cache
        self.spec = spec

    def create_module(self, spec):
        # Returning None means to create the (uninitialized) module
        # in the same way as normal.
        #
        # (But we could return e.g. a subclass of Module if we wanted.)
        return None

    def exec_module(self, module):
        if world.rank == 0:
            # Load from file and store in cache:
            code = self.spec.loader.get_code(module.__name__)
            metadata = (self.spec.submodule_search_locations, self.spec.origin)
            self.module_cache[module.__name__] = (metadata, code)
            # We could execute the default mechanism to load the module here.
            # Instead we load from cache using our own loader, like on the
            # other cores.

        return self.load_from_cache(module)

    def load_from_cache(self, module):
        metadata, code = self.module_cache[module.__name__]
        origin = metadata[1]
        module.__file__ = origin
        # __package__, __path__, __cached__?
        module.__loader__ = self
        sys.modules[module.__name__] = module
        exec(code, module.__dict__)
        return module

    def __str__(self):
        return ('<{} for {}:{} [{} modules cached]>'
                .format(self.__class__.__name__,
                        self.spec.name, self.spec.origin,
                        len(self.module_cache)))


class BroadcastImporter:
    def __init__(self):
        self.module_cache = {}
        self.cached_modules = []

    def find_spec(self, fullname, path=None, target=None):
        if world.rank == 0:
            spec = PathFinder.find_spec(fullname, path, target)
            if spec is None:
                return None

            if spec.loader is None:
                return None

            code = spec.loader.get_code(fullname)
            if code is None:  # C extensions
                return None

            loader = BroadcastLoader(spec, self.module_cache)
            assert fullname == spec.name

            searchloc = spec.submodule_search_locations
            spec = ModuleSpec(fullname, loader, origin=spec.origin,
                              is_package=searchloc is not None)
            if searchloc is not None:
                spec.submodule_search_locations += searchloc
            return spec
        else:
            if fullname not in self.module_cache:
                # Could this in principle interfere with builtin imports?
                return PathFinder.find_spec(fullname, path, target)

            searchloc, origin = self.module_cache[fullname][0]
            loader = BroadcastLoader(None, self.module_cache)
            spec = ModuleSpec(fullname, loader, origin=origin,
                              is_package=searchloc is not None)
            if searchloc is not None:
                spec.submodule_search_locations += searchloc
            loader.spec = spec  # XXX loader.loader is still None
            return spec

    def broadcast(self):
        if world.size == 1:
            return
        if world.rank == 0:
            # print('bcast {} modules'.format(len(self.module_cache)))
            marshal_broadcast(self.module_cache)
        else:
            self.module_cache = marshal_broadcast(None)
            # print('recv {} modules'.format(len(self.module_cache)))

    def enable(self):
        if world is None:
            return

        # There is the question of whether we lose anything by inserting
        # ourselves further on in the meta_path list.  Maybe not, and maybe
        # that is a less violent act.
        sys.meta_path.insert(0, self)
        if world.rank != 0:
            self.broadcast()

    def disable(self):
        if world is None:
            return

        if world.rank == 0:
            self.broadcast()
        self.cached_modules += self.module_cache.keys()
        self.module_cache = {}
        myself = sys.meta_path.pop(0)
        assert myself is self

    def __enter__(self):
        self.enable()

    def __exit__(self, *args):
        self.disable()


broadcast_imports = BroadcastImporter()



import pytest
import numpy as np
from ase import Atoms
from ase.lattice.hexagonal import Hexagonal
from gpaw import GPAW, FermiDirac, PW
from gpaw.mpi import world
from gpaw.test import findpeak, equal
from gpaw.response.bse import BSE
from gpaw.response.df import read_response_function


@pytest.mark.response
def test_response_bse_MoS2_cut(in_tmp_dir, scalapack):
    calc = GPAW(mode=PW(180),
                xc='PBE',
                nbands='nao',
                setups={'Mo': '6'},
                occupations=FermiDirac(0.001),
                convergence={'bands': -5},
                kpts=(5, 5, 1))

    a = 3.1604
    c = 10.0

    cell = Hexagonal(symbol='Mo',
                     latticeconstant={'a': a, 'c': c}).get_cell()
    layer = Atoms(symbols='MoS2', cell=cell, pbc=(1, 1, 0),
                  scaled_positions=[(0, 0, 0),
                                    (2 / 3, 1 / 3, 0.3),
                                    (2 / 3, 1 / 3, -0.3)])

    pos = layer.get_positions()
    pos[1][2] = pos[0][2] + 3.172 / 2
    pos[2][2] = pos[0][2] - 3.172 / 2
    layer.set_positions(pos)
    layer.set_pbc([True, True, False])
    layer.center(axis=2)
    layer.calc = calc
    layer.get_potential_energy()
    calc.write('MoS2.gpw', mode='all')

    bse = BSE('MoS2.gpw',
              spinors=True,
              ecut=10,
              valence_bands=[8],
              conduction_bands=[9],
              eshift=0.8,
              nbands=15,
              write_h=False,
              write_v=False,
              wfile=None,
              mode='BSE',
              truncation='2D')

    outw_w, outalpha_w = bse.get_polarizability(write_eig=None,
                                                eta=0.02,
                                                w_w=np.linspace(0., 5., 5001))
    world.barrier()
    w_w, alphareal_w, alphaimag_w = read_response_function('pol_bse.csv')

    # Check consistency with written results
    assert np.allclose(outw_w, w_w, atol=1e-5, rtol=1e-4)
    assert np.allclose(outalpha_w.real, alphareal_w, atol=1e-5, rtol=1e-4)
    assert np.allclose(outalpha_w.imag, alphaimag_w, atol=1e-5, rtol=1e-4)

    w0, I0 = findpeak(w_w[:1100], alphaimag_w[:1100])
    w1, I1 = findpeak(w_w[1100:1300], alphaimag_w[1100:1300])
    w1 += 1.1

    equal(w0, 0.58, 0.01)
    equal(I0, 38.8, 0.35)
    equal(w1, 2.22, 0.01)
    equal(I1, 6.3, 0.35)



from __future__ import annotations

import importlib
import os
from types import ModuleType, SimpleNamespace
from typing import Any, Union

import numpy as np
from ase import Atoms
from ase.calculators.calculator import kpts2sizeandoffsets
from ase.units import Bohr

from gpaw.core import UGDesc
from gpaw.core.atom_arrays import (AtomArrays, AtomArraysLayout,
                                   AtomDistribution)
from gpaw.core.domain import Domain
from gpaw.gpu.mpi import CuPyMPI
from gpaw.mixer import MixerWrapper, get_mixer_from_keywords
from gpaw.mpi import (MPIComm, Parallelization, serial_comm, synchronize_atoms,
                      world)
from gpaw.new import cached_property, prod
from gpaw.new.basis import create_basis
from gpaw.new.brillouin import BZPoints, MonkhorstPackKPoints
from gpaw.new.density import Density
from gpaw.new.ibzwfs import IBZWaveFunctions
from gpaw.new.input_parameters import InputParameters
from gpaw.new.scf import SCFLoop
from gpaw.new.smearing import OccupationNumberCalculator
from gpaw.new.symmetry import create_symmetries_object
from gpaw.new.xc import create_functional
from gpaw.setup import Setups
from gpaw.typing import Array2D, ArrayLike1D, ArrayLike2D
from gpaw.utilities.gpts import get_number_of_grid_points
from gpaw.xc import XC
from gpaw.new.c import GPU_AWARE_MPI


def builder(atoms: Atoms,
            params: dict[str, Any] | InputParameters,
            comm=None) -> DFTComponentsBuilder:
    """Create DFT-components builder.

    * pw
    * lcao
    * fd
    * tb
    * atom
    """
    if isinstance(params, dict):
        params = InputParameters(params)

    mode = params.mode.copy()
    name = mode.pop('name')
    assert name in {'pw', 'lcao', 'fd', 'tb', 'atom'}
    mod = importlib.import_module(f'gpaw.new.{name}.builder')
    name = name.title() if name == 'atom' else name.upper()
    return getattr(mod, f'{name}DFTComponentsBuilder')(
        atoms, params, comm=comm or world, **mode)


class DFTComponentsBuilder:
    def __init__(self,
                 atoms: Atoms,
                 params: InputParameters,
                 *,
                 comm):

        self.atoms = atoms.copy()
        self.mode = params.mode['name']
        self.params = params

        parallel = params.parallel

        synchronize_atoms(atoms, comm)
        self.check_cell(atoms.cell)

        self.initial_magmom_av, self.ncomponents = normalize_initial_magmoms(
            atoms, params.magmoms, params.spinpol or params.hund)

        self.soc = params.soc
        self.nspins = self.ncomponents % 3
        self.spin_degeneracy = self.ncomponents % 2 + 1

        if isinstance(params.xc, (dict, str)):
            self._xc = XC(params.xc, collinear=(self.ncomponents < 4))
        else:
            self._xc = params.xc

        self.setups = Setups(atoms.numbers,
                             params.setups,
                             params.basis,
                             self._xc.get_setup_name(),
                             world=comm)

        if params.hund:
            c = params.charge / len(atoms)
            for a, setup in enumerate(self.setups):
                self.initial_magmom_av[a, 2] = setup.get_hunds_rule_moment(c)

        symmetries = create_symmetries_object(atoms,
                                              self.setups.id_a,
                                              self.initial_magmom_av,
                                              params.symmetry)
        assert not (self.ncomponents == 4 and len(symmetries) > 1)
        bz = create_kpts(params.kpts, atoms)
        self.ibz = symmetries.reduce(bz, strict=False)

        d = parallel.get('domain', None)
        k = parallel.get('kpt', None)
        b = parallel.get('band', None)
        self.communicators = create_communicators(comm, len(self.ibz),
                                                  d, k, b, self.xp)

        if self.mode == 'fd':
            pass  # filter = create_fourier_filter(grid)
            # setups = setups.filter(filter)

        self.nelectrons = self.setups.nvalence - params.charge

        self.nbands = calculate_number_of_bands(params.nbands,
                                                self.setups,
                                                params.charge,
                                                self.initial_magmom_av,
                                                self.mode == 'lcao')
        if self.ncomponents == 4:
            self.nbands *= 2

        self.dtype = params.dtype
        if self.dtype is None:
            if self.ibz.bz.gamma_only:
                self.dtype = float
            else:
                self.dtype = complex
        elif not self.ibz.bz.gamma_only and self.dtype != complex:
            raise ValueError('Can not use dtype=float for non gamma-point '
                             'calculation')

        self.grid, self.fine_grid = self.create_uniform_grids()

        self.fracpos_ac = self.atoms.get_scaled_positions()
        self.fracpos_ac %= 1
        self.fracpos_ac %= 1

        self.xc = self.create_xc_functional()

        self.interpolation_desc: Domain
        self.electrostatic_potential_desc: Domain
        self.atomdist: AtomDistribution

    def __repr__(self):
        return f'{self.__class__.__name__}({self.atoms}, {self.params})'

    def create_uniform_grids(self):
        raise NotImplementedError

    def create_xc_functional(self):
        return create_functional(self._xc,
                                 self.fine_grid)

    def check_cell(self, cell):
        number_of_lattice_vectors = cell.rank
        if number_of_lattice_vectors < 3:
            raise ValueError(
                'GPAW requires 3 lattice vectors.  '
                f'Your system has {number_of_lattice_vectors}.')

    @cached_property
    def wf_desc(self) -> Domain:
        return self.create_wf_description()

    @cached_property
    def xp(self) -> ModuleType:
        """Array module: Numpy or Cupy."""
        if self.params.parallel['gpu']:
            from gpaw.gpu import cupy, cupy_is_fake
            assert not cupy_is_fake or os.environ.get('GPAW_CPUPY')
            return cupy
        return np

    def create_wf_description(self) -> Domain:
        raise NotImplementedError

    def get_pseudo_core_densities(self):
        raise NotImplementedError

    def get_pseudo_core_ked(self):
        raise NotImplementedError

    def create_basis_set(self):
        return create_basis(self.ibz,
                            self.ncomponents % 3,
                            self.atoms.pbc,
                            self.grid,
                            self.setups,
                            self.dtype,
                            self.fracpos_ac,
                            self.communicators['w'],
                            self.communicators['k'],
                            self.communicators['b'])

    def density_from_superposition(self, basis_set):
        return Density.from_superposition(
            grid=self.grid,
            nct_aX=self.get_pseudo_core_densities(),
            tauct_aX=self.get_pseudo_core_ked(),
            atomdist=self.atomdist,
            setups=self.setups,
            basis_set=basis_set,
            magmom_av=self.initial_magmom_av,
            ncomponents=self.ncomponents,
            charge=self.params.charge,
            hund=self.params.hund,
            mgga=self.xc.type == 'MGGA')

    def create_occupation_number_calculator(self):
        return OccupationNumberCalculator(
            self.params.occupations,
            self.atoms.pbc,
            self.ibz,
            self.nbands,
            self.communicators,
            self.initial_magmom_av.sum(0),
            self.ncomponents,
            np.linalg.inv(self.atoms.cell.complete()).T)

    def create_hamiltonian_operator(self):
        raise NotImplementedError

    def create_eigensolver(self, hamiltonian):
        raise NotImplementedError

    def create_scf_loop(self):
        hamiltonian = self.create_hamiltonian_operator()
        occ_calc = self.create_occupation_number_calculator()
        eigensolver = self.create_eigensolver(hamiltonian)

        mixer = MixerWrapper(
            get_mixer_from_keywords(self.atoms.pbc.any(),
                                    self.ncomponents, **self.params.mixer),
            self.ncomponents,
            self.grid._gd,
            world=self.communicators['w'])

        return SCFLoop(hamiltonian, occ_calc,
                       eigensolver, mixer, self.communicators['w'],
                       {key: value
                        for key, value in self.params.convergence.items()
                        if key != 'bands'},
                       self.params.maxiter)

    def read_ibz_wave_functions(self, reader):
        raise NotImplementedError

    def create_potential_calculator(self):
        raise NotImplementedError

    def read_wavefunction_values(self,
                                 reader,
                                 ibzwfs: IBZWaveFunctions) -> None:
        """Read eigenvalues, occuptions and projections and fermi levels.

        The values are read using reader and set as the appropriate properties
        of (the already instantiated) wavefunctions contained in ibzwfs
        """
        ha = reader.ha

        eig_skn = reader.wave_functions.eigenvalues
        occ_skn = reader.wave_functions.occupations
        if self.communicators['d'].rank == 0:
            P_sknI = reader.wave_functions.projections
            P_sknI = P_sknI.astype(ibzwfs.dtype)
        else:
            P_sknI = None

        for wfs in ibzwfs:
            wfs._eig_n = eig_skn[wfs.spin, wfs.k] / ha
            wfs._occ_n = occ_skn[wfs.spin, wfs.k]
            layout = AtomArraysLayout([(setup.ni,) for setup in self.setups],
                                      atomdist=self.atomdist,
                                      dtype=self.dtype)
            data = None
            if self.ncomponents < 4:
                wfs._P_ani = AtomArrays(layout, dims=(self.nbands,))
                if P_sknI is not None:
                    data = P_sknI[wfs.spin, wfs.k]
            else:
                wfs._P_ani = AtomArrays(layout, dims=(self.nbands, 2))
                if P_sknI is not None:
                    data = P_sknI[wfs.k]
            wfs._P_ani.scatter_from(data)

        try:
            ibzwfs.fermi_levels = reader.wave_functions.fermi_levels / ha
        except AttributeError:
            # old gpw-file
            ibzwfs.fermi_levels = np.array(
                [reader.occupations.fermilevel / ha])


def create_communicators(comm: MPIComm = None,
                         nibzkpts: int = 1,
                         domain: Union[int, tuple[int, int, int]] = None,
                         kpt: int = None,
                         band: int = None,
                         xp: ModuleType = np) -> dict[str, MPIComm]:
    parallelization = Parallelization(comm or world, nibzkpts)
    if domain is not None and not isinstance(domain, int):
        domain = prod(domain)
    parallelization.set(kpt=kpt,
                        domain=domain,
                        band=band)
    comms = parallelization.build_communicators()
    comms['w'] = comm

    # We replace size=1 MPI communications with serial_comm so that
    # serial_comm.sum(<cupy-array>) works: XXX
    comms = {key: comm if comm.size > 1 else serial_comm
             for key, comm in comms.items()}

    if xp is not np and not GPU_AWARE_MPI:
        comms = {key: CuPyMPI(comm) for key, comm in comms.items()}

    return comms


def create_fourier_filter(grid):
    gamma = 1.6

    h = ((grid.icell**2).sum(1)**-0.5 / grid.size).max()

    def filter(rgd, rcut, f_r, l=0):
        gcut = np.pi / h - 2 / rcut / gamma
        ftmp = rgd.filter(f_r, rcut * gamma, gcut, l)
        f_r[:] = ftmp[:len(f_r)]

    return filter


def normalize_initial_magmoms(
        atoms: Atoms,
        magmoms: ArrayLike2D | ArrayLike1D | float | None = None,
        force_spinpol_calculation: bool = False) -> tuple[Array2D, int]:
    """Convert magnetic moments to (natoms, 3)-shaped array.

    Also return number of wave function components (1, 2 or 4).

    >>> h = Atoms('H', magmoms=[1])
    >>> normalize_initial_magmoms(h)
    (array([[0., 0., 1.]]), 2)
    >>> normalize_initial_magmoms(h, [[1, 0, 0]])
    (array([[1., 0., 0.]]), 4)
    """
    magmom_av = np.zeros((len(atoms), 3))
    ncomponents = 2

    if magmoms is None:
        magmom_av[:, 2] = atoms.get_initial_magnetic_moments()
    elif isinstance(magmoms, float):
        magmom_av[:, 2] = magmoms
    else:
        magmoms = np.asarray(magmoms)
        if magmoms.ndim == 1:
            magmom_av[:, 2] = magmoms
        else:
            magmom_av[:] = magmoms
            ncomponents = 4

    if (ncomponents == 2 and
        not force_spinpol_calculation and
        not magmom_av[:, 2].any()):
        ncomponents = 1

    return magmom_av, ncomponents


def create_kpts(kpts: dict[str, Any], atoms: Atoms) -> BZPoints:
    if 'kpts' in kpts:
        assert len(kpts) == 1, kpts
        return BZPoints(kpts['kpts'])
    size, offset = kpts2sizeandoffsets(**kpts, atoms=atoms)
    return MonkhorstPackKPoints(size, offset)


def calculate_number_of_bands(nbands: int | str | None,
                              setups: Setups,
                              charge: float,
                              initial_magmom_av: Array2D,
                              is_lcao: bool) -> int:
    nao = setups.nao
    nvalence = setups.nvalence - charge
    M = np.linalg.norm(initial_magmom_av.sum(0))

    orbital_free = any(setup.orbital_free for setup in setups)
    if orbital_free:
        return 1

    if isinstance(nbands, str):
        if nbands == 'nao':
            N = nao
        elif nbands[-1] == '%':
            cfgbands = (nvalence + M) / 2
            N = int(np.ceil(float(nbands[:-1]) / 100 * cfgbands))
        else:
            raise ValueError('Integer expected: Only use a string '
                             'if giving a percentage of occupied bands')
    elif nbands is None:
        # Number of bound partial waves:
        nbandsmax = sum(setup.get_default_nbands()
                        for setup in setups)
        N = int(np.ceil((1.2 * (nvalence + M) / 2))) + 4
        N = min(N, nbandsmax)
        if is_lcao and N > nao:
            N = nao
    elif nbands <= 0:
        N = max(1, int(nvalence + M + 0.5) // 2 + (-nbands))
    else:
        N = nbands

    if N > nao and is_lcao:
        raise ValueError('Too many bands for LCAO calculation: '
                         f'{nbands}%d bands and only {nao} atomic orbitals!')

    if nvalence < 0:
        raise ValueError(
            f'Charge {charge} is not possible - not enough valence electrons')

    if nvalence > 2 * N:
        raise ValueError(
            f'Too few bands!  Electrons: {nvalence}, bands: {nbands}')

    return N


def create_uniform_grid(mode: str,
                        gpts,
                        cell,
                        pbc,
                        symmetry,
                        h: float = None,
                        interpolation: str = None,
                        ecut: float = None,
                        comm: MPIComm = serial_comm) -> UGDesc:
    """Create grid in a backwards compatible way."""
    cell = cell / Bohr
    if h is not None:
        h /= Bohr

    realspace = (mode != 'pw' and interpolation != 'fft')
    if not realspace:
        pbc = (True, True, True)

    if gpts is not None:
        size = gpts
    else:
        modeobj = SimpleNamespace(name=mode, ecut=ecut)
        size = get_number_of_grid_points(cell, h, modeobj, realspace,
                                         symmetry.symmetry)
    return UGDesc(cell=cell, pbc=pbc, size=size, comm=comm)



"""Omega matrix for functionals with Hartree-Fock exchange.

"""
from math import sqrt

from ase.units import Hartree
from ase.utils.timing import Timer
from ase.utils import IOContext
import numpy as np
from numpy.linalg import inv
from scipy.linalg import eigh

from gpaw import debug
import gpaw.mpi as mpi
from gpaw.lrtddft.omega_matrix import OmegaMatrix
from gpaw.pair_density import PairDensity
from gpaw.helmholtz import HelmholtzSolver
from gpaw.utilities.blas import mmm


class ApmB(OmegaMatrix):

    """Omega matrix for functionals with Hartree-Fock exchange.

    """

    def get_full(self):

        hybrid = ((self.xc is not None) and
                  hasattr(self.xc, 'hybrid') and
                  (self.xc.hybrid > 0.0))
        if self.fullkss.npspins < 2 and hybrid:
            raise RuntimeError('Does not work spin-unpolarized ' +
                               'with hybrids (use nspins=2)')

        if hasattr(self.xc, 'rsf') and (self.xc.rsf == 'Yukawa'):
            self.screened_poissonsolver = HelmholtzSolver(
                k2=-self.xc.omega**2, eps=1e-11, nn=3)
            self.screened_poissonsolver.set_grid_descriptor(self.gd)
        self.paw.timer.start('ApmB RPA')
        self.ApB = self.Om
        self.AmB = self.get_rpa()
        self.paw.timer.stop()

        if self.xc is not None:
            self.paw.timer.start('ApmB XC')
            self.get_xc()  # inherited from OmegaMatrix
            self.paw.timer.stop()

    def get_rpa(self):
        """Calculate RPA and Hartree-fock part of the A+-B matrices."""

        # shorthands
        kss = self.fullkss
        finegrid = self.finegrid
        yukawa = hasattr(self.xc, 'rsf') and (self.xc.rsf == 'Yukawa')

        # calculate omega matrix
        nij = len(kss)
        self.log('RPAhyb', nij, 'transitions')

        AmB = np.zeros((nij, nij))
        ApB = self.ApB

        # storage place for Coulomb integrals
        integrals = {}
        if yukawa:
            rsf_integrals = {}
        # setup things for IVOs
        if (hasattr(self.xc, 'excitation') and
           (self.xc.excitation is not None or self.xc.excited != 0)):
            sin_tri_weight = 1
            if self.xc.excitation is not None:
                ex_type = self.xc.excitation.lower()
                if ex_type == 'singlet':
                    sin_tri_weight = 2
                elif ex_type == 'triplet':
                    sin_tri_weight = 0
            homo = int(self.paw.get_number_of_electrons() // 2)
            ivo_l = homo - self.xc.excited - 1
        else:
            ivo_l = None

        for ij in range(nij):
            self.log('RPAhyb kss[' + '%d' % ij + ']=', kss[ij])

            timer = Timer()
            timer.start('init')
            timer2 = Timer()

            # smooth density including compensation charges
            timer2.start('with_compensation_charges 0')
            rhot_p = kss[ij].with_compensation_charges(
                finegrid != 0)
            timer2.stop()

            # integrate with 1/|r_1-r_2|
            timer2.start('poisson')
            phit_p = np.zeros(rhot_p.shape, rhot_p.dtype)
            self.poisson.solve(phit_p, rhot_p, charge=None)
            timer2.stop()

            timer.stop()
            t0 = timer.get_time('init')
            timer.start(ij)

            if finegrid == 1:
                rhot = kss[ij].with_compensation_charges()
                phit = self.gd.zeros()
                self.restrict(phit_p, phit)
            else:
                phit = phit_p
                rhot = rhot_p

            for kq in range(ij, nij):
                if kq != ij:
                    # smooth density including compensation charges
                    timer2.start('kq with_compensation_charges')
                    rhot = kss[kq].with_compensation_charges(
                        finegrid == 2)
                    timer2.stop()
                pre = self.weight_Kijkq(ij, kq)

                timer2.start('integrate')
                I = self.Coulomb_integral_kss(kss[ij], kss[kq], phit, rhot)
                if kss[ij].spin == kss[kq].spin:
                    name = self.Coulomb_integral_name(kss[ij].i, kss[ij].j,
                                                      kss[kq].i, kss[kq].j,
                                                      kss[ij].spin)
                    integrals[name] = I
                ApB[ij, kq] = pre * I
                timer2.stop()

                if ij == kq:
                    epsij = kss[ij].get_energy() / kss[ij].get_weight()
                    AmB[ij, kq] += epsij
                    ApB[ij, kq] += epsij

            timer.stop()
# timer2.write()
            if ij < (nij - 1):
                self.log('RPAhyb estimated time left',
                         self.time_left(timer, t0, ij, nij))

        # add HF parts and apply symmetry
        if hasattr(self.xc, 'hybrid'):
            weight = self.xc.hybrid
        else:
            weight = 0.0
        for ij in range(nij):
            self.log('HF kss[' + '%d' % ij + ']')
            timer = Timer()
            timer.start('init')
            timer.stop()
            t0 = timer.get_time('init')
            timer.start(ij)

            i = kss[ij].i
            j = kss[ij].j
            s = kss[ij].spin
            for kq in range(ij, nij):
                if kss[ij].pspin == kss[kq].pspin:
                    k = kss[kq].i
                    q = kss[kq].j
                    ikjq = self.Coulomb_integral_ijkq(i, k, j, q, s, integrals)
                    iqkj = self.Coulomb_integral_ijkq(i, q, k, j, s, integrals)
                    if yukawa:  # Yukawa integrals might be caches
                        ikjq -= self.Coulomb_integral_ijkq(
                            i, k, j, q, s, rsf_integrals, yukawa)
                        iqkj -= self.Coulomb_integral_ijkq(
                            i, q, k, j, s, rsf_integrals, yukawa)
                    ApB[ij, kq] -= weight * (ikjq + iqkj)
                    AmB[ij, kq] -= weight * (ikjq - iqkj)

                ApB[kq, ij] = ApB[ij, kq]
                AmB[kq, ij] = AmB[ij, kq]

            timer.stop()
            if ij < (nij - 1):
                self.log('HF estimated time left',
                         self.time_left(timer, t0, ij, nij))

        if ivo_l is not None:
            # IVO RPA after Berman, Kaldor, Chem. Phys. 43 (3) 1979
            # doi: 10.1016/0301-0104(79)85205-2
            l = ivo_l
            for ij in range(nij):
                i = kss[ij].i
                j = kss[ij].j
                s = kss[ij].spin
                for kq in range(ij, nij):
                    if kss[kq].i == i and kss[ij].pspin == kss[kq].pspin:
                        k = kss[kq].i
                        q = kss[kq].j
                        jqll = self.Coulomb_integral_ijkq(j, q, l, l, s,
                                                          integrals)
                        jllq = self.Coulomb_integral_ijkq(j, l, l, q, s,
                                                          integrals)
                        if yukawa:
                            jqll -= self.Coulomb_integral_ijkq(j, q, l, l, s,
                                                               rsf_integrals,
                                                               yukawa)
                            jllq -= self.Coulomb_integral_ijkq(j, l, l, q, s,
                                                               rsf_integrals,
                                                               yukawa)
                        jllq *= sin_tri_weight
                        ApB[ij, kq] += weight * (jqll - jllq)
                        AmB[ij, kq] += weight * (jqll - jllq)
                        ApB[kq, ij] = ApB[ij, kq]
                        AmB[kq, ij] = AmB[ij, kq]
        return AmB

    def Coulomb_integral_name(self, i, j, k, l, spin):
        """return a unique name considering the Coulomb integral
        symmetry"""
        def ij_name(i, j):
            return str(max(i, j)) + ' ' + str(min(i, j))

        # maximal gives the first
        if max(i, j) >= max(k, l):
            base = ij_name(i, j) + ' ' + ij_name(k, l)
        else:
            base = ij_name(k, l) + ' ' + ij_name(i, j)
        return base + ' ' + str(spin)

    def Coulomb_integral_ijkq(self, i, j, k, q, spin, integrals,
                              yukawa=False):
        name = self.Coulomb_integral_name(i, j, k, q, spin)
        if name in integrals:
            return integrals[name]

        # create the Kohn-Sham singles
        kss_ij = PairDensity(self.paw)
        kss_ij.initialize(self.paw.wfs.kpt_u[spin], i, j)
        kss_kq = PairDensity(self.paw)
        kss_kq.initialize(self.paw.wfs.kpt_u[spin], k, q)

        rhot_p = kss_ij.with_compensation_charges(
            self.finegrid != 0)
        phit_p = np.zeros(rhot_p.shape, rhot_p.dtype)
        if yukawa:
            self.screened_poissonsolver.solve(phit_p, rhot_p, charge=None)
        else:
            self.poisson.solve(phit_p, rhot_p, charge=None)

        if self.finegrid == 1:
            phit = self.gd.zeros()
            self.restrict(phit_p, phit)
        else:
            phit = phit_p

        rhot = kss_kq.with_compensation_charges(
            self.finegrid == 2)

        integrals[name] = self.Coulomb_integral_kss(kss_ij, kss_kq,
                                                    phit, rhot,
                                                    yukawa=yukawa)
        return integrals[name]

    def timestring(self, t):
        ti = int(t + .5)
        td = int(ti // 86400)
        st = ''
        if td > 0:
            st += '%d' % td + 'd'
            ti -= td * 86400
        th = int(ti // 3600)
        if th > 0:
            st += '%d' % th + 'h'
            ti -= th * 3600
        tm = int(ti // 60)
        if tm > 0:
            st += '%d' % tm + 'm'
            ti -= tm * 60
        st += '%d' % ti + 's'
        return st

    def mapAB(self, restrict={}):
        """Map A+B, A-B matrices according to constraints."""

        map, self.kss = self.get_map(restrict)
        if map is None:
            ApB = self.ApB.copy()
            AmB = self.AmB.copy()
        else:
            nij = len(self.kss)
            ApB = np.empty((nij, nij))
            AmB = np.empty((nij, nij))
            for ij in range(nij):
                for kq in range(nij):
                    ApB[ij, kq] = self.ApB[map[ij], map[kq]]
                    AmB[ij, kq] = self.AmB[map[ij], map[kq]]

        return ApB, AmB

    def diagonalize(self, restrict={}, TDA=False):
        """Evaluate Eigenvectors and Eigenvalues"""

        ApB, AmB = self.mapAB(restrict)
        nij = len(self.kss)

        if TDA:
            # Tamm-Dancoff approximation (B=0)
            eigenvalues, evecs = eigh(0.5 * (ApB + AmB))
            self.eigenvectors = evecs.T
            self.eigenvalues = eigenvalues ** 2
        else:
            # the occupation matrix
            C = np.empty((nij,))
            for ij in range(nij):
                C[ij] = 1. / self.kss[ij].fij

            S = C * inv(AmB) * C
            S = sqrt_matrix(inv(S).copy())

            # get Omega matrix
            M = np.zeros(ApB.shape)
            mmm(1.0, S, 'N', ApB, 'N', 0.0, M)
            self.eigenvectors = np.zeros(ApB.shape)
            mmm(1.0, M, 'N', S, 'N', 0.0, self.eigenvectors)

            self.eigenvalues, self.eigenvectors.T[:] = eigh(self.eigenvectors)

    def read(self, filename=None, fh=None):
        """Read myself from a file"""
        if mpi.rank == 0:
            with IOContext() as io:
                if fh is None:
                    fd = io.openfile(filename, 'r')
                else:
                    fd = fh
                fd.readline()
                nij = int(fd.readline())
                ApB = np.zeros((nij, nij))
                for ij in range(nij):
                    l = fd.readline().split()
                    for kq in range(ij, nij):
                        ApB[ij, kq] = float(l[kq - ij])
                        ApB[kq, ij] = ApB[ij, kq]
                self.ApB = ApB

                fd.readline()
                nij = int(fd.readline())
                AmB = np.zeros((nij, nij))
                for ij in range(nij):
                    l = fd.readline().split()
                    for kq in range(ij, nij):
                        AmB[ij, kq] = float(l[kq - ij])
                        AmB[kq, ij] = AmB[ij, kq]
                self.AmB = AmB

    def weight_Kijkq(self, ij, kq):
        """weight for the coupling matrix terms"""
        return 2.

    def write(self, filename=None, fh=None):
        """Write current state to a file."""
        if mpi.rank == 0:
            with IOContext() as io:
                if fh is None:
                    fd = io.openfile(filename, 'r')
                else:
                    fd = fh
                fd.write('# A+B\n')
                nij = len(self.fullkss)
                fd.write('%d\n' % nij)
                for ij in range(nij):
                    for kq in range(ij, nij):
                        fd.write(' %g' % self.ApB[ij, kq])
                    fd.write('\n')

                fd.write('# A-B\n')
                nij = len(self.fullkss)
                fd.write('%d\n' % nij)
                for ij in range(nij):
                    for kq in range(ij, nij):
                        fd.write(' %g' % self.AmB[ij, kq])
                    fd.write('\n')

    def __str__(self):
        string = '<ApmB> '
        if hasattr(self, 'eigenvalues'):
            string += 'dimension ' + ('%d' % len(self.eigenvalues))
            string += '\neigenvalues: '
            for ev in self.eigenvalues:
                string += ' ' + ('%f' % (sqrt(ev) * Hartree))
        return string


def sqrt_matrix(a, preserve=False):
    """Get the sqrt of a symmetric matrix a (diagonalize is used).
    The matrix is kept if preserve=True, a=sqrt(a) otherwise."""
    n = len(a)
    if debug:
        assert a.flags.contiguous
        assert a.dtype == float
        assert a.shape == (n, n)
    if preserve:
        b = a.copy()
    else:
        b = a

    # diagonalize to get the form b = Z * D * Z^T
    # where D is diagonal
    D = np.empty((n,))
    D, b.T[:] = eigh(b, lower=True)
    ZT = b.copy()
    Z = np.transpose(b)

    # c = Z * sqrt(D)
    c = Z * np.sqrt(D)

    # sqrt(b) = c * Z^T
    mmm(1.0, np.ascontiguousarray(c), 'N', ZT, 'N', 0.0, b)
    return b



import pytest
from ase import Atoms
from gpaw import GPAW, FermiDirac
from gpaw.lrtddft import LrTDDFT


@pytest.mark.lrtddft
def test_digonalize():
    """Test selection at diagonalization stage"""
    atoms = Atoms('O')
    atoms.cell = [3, 4, 5]
    atoms.center()

    atoms.calc = GPAW(mode='fd',
                      occupations=FermiDirac(width=0.1),
                      nbands=5)
    atoms.get_potential_energy()

    lr = LrTDDFT(atoms.calc)

    # all
    lr.diagonalize()
    assert len(lr) == 10

    lr.diagonalize(restrict={'istart': 3})
    assert len(lr) == 1

    lr.diagonalize(restrict={'jend': 1})
    assert len(lr) == 1

    lr.diagonalize(restrict={'eps': 1.5})
    assert len(lr) == 2

    lr.diagonalize(restrict={'energy_range': 1})
    assert len(lr) == 3



"""Module for Numpy array diagonalization with Scipy/Scalapack."""
import numpy as np
from scipy.linalg import eigh

from gpaw.blacs import BlacsGrid, Redistributor


class ScipyDiagonalizer:
    """Diagonalizer class that uses scipy.linalg.eigh.

    The ScipyDiagonalizer wraps scipy.linalg.eigh to solve a
    (generalized) eigenproblem on one core.
    """

    def diagonalize(self, A, B, eps, is_master, debug):
        """Solves the eigenproblem A @ x = eps [B] @ x.

        The problem is solved inplace, so when done, A has the eigenvectors as
        columns and eps has the eigenvalues.
        B is overwritten for potential increase in performance.

        Parameters
        ----------
        A : Numpy array
            Left-hand matrix of the eigenproblem. After running, the
            eigenvectors are the column vectors of this array.
        B : Numpy array
            Right-hand overlap matrix of the eigenproblem.
        eps : Numpy array
            1D vector containing the eigenvalues of the solved eigenproblem.
        is_master : bool
            A boolean to mark which rank to perform the diagonalization on.
            Since Scipy's diagonalizer is not MPI-parallelized, the
            diagonalization needs (necessarily) to be done on the rank which
            has the arrays.
        debug : bool
            Flag to check for finiteness when running in debug mode.
        """
        if is_master:
            eps[:], A[:] = eigh(
                A, B, lower=True, check_finite=debug, overwrite_b=True)


class ScalapackDiagonalizer:
    """Diagonalizer class that uses general_diagonalize_dc.

    The ScalapackDiagonalizer wraps general_diagonalize_dc to solve a
    (generalized) eigenproblem on one core.
    """

    def __init__(
        self,
        arraysize,
        grid_nrows,
        grid_ncols,
        *,
        dtype,
        scalapack_communicator,
        blocksize=64):
        """Initialize grids, communicators, redistributors.

        Parameters
        ----------
        arraysize : int
            The side length of the square matrix to diagonalize.
        grid_nrows : int
            Number of rows in the BLACS grid.
        grid_ncols : int
            Number of columns in the BLACS grid.
        dtype : type
            `float` or `complex`, the datatype of the eigenproblem.
        scalapack_communicator : MPICommunicator
            The communicator object over which Scalapack diagonalizes.
        blocksize : int, optional
            The block size in the 2D block cyclic data distribution.
            The default value of 64 is universally good.
        """
        self.arraysize = arraysize
        self.blocksize = blocksize
        self.dtype = dtype

        self.blacsgrid = BlacsGrid(
            scalapack_communicator, grid_nrows, grid_ncols)
        self.distributed_descriptor = self.blacsgrid.new_descriptor(
            arraysize, arraysize, blocksize, blocksize)
        self.head_rank_descriptor = self.blacsgrid.new_descriptor(
            arraysize, arraysize, arraysize, arraysize)

        self.head_to_all_redistributor = Redistributor(
            self.blacsgrid.comm,
            self.head_rank_descriptor,
            self.distributed_descriptor)

        self.all_to_head_redistributor = Redistributor(
            self.blacsgrid.comm,
            self.distributed_descriptor,
            self.head_rank_descriptor)

    def diagonalize(self, A, B, eps, is_master, debug):
        """Solves the eigenproblem A @ x = eps [B] @ x.

        The problem is solved inplace, so when done, A has the eigenvectors
        as columns and eps has the eigenvalues.

        Parameters
        ----------
        A : Numpy array
            Left-hand matrix of the eigenproblem. After running, the
            eigenvectors are the column vectors of this array.
        B : Numpy array
            Right-hand overlap matrix of the eigenproblem.
        eps : Numpy array
            1D vector containing the eigenvalues of the solved eigenproblem.
        is_master : bool
            A boolean to mark which rank to perform the diagonalization on.
            Used to know which ranks to redistribute the Numpy arrays to/from.
        debug : bool
            Flag to check for finiteness when running in debug mode.
        """
        Asc_MM = self.head_rank_descriptor.zeros(dtype=self.dtype)
        Bsc_MM = self.head_rank_descriptor.zeros(dtype=self.dtype)
        vec_MM = self.head_rank_descriptor.zeros(dtype=self.dtype)

        Asc_mm = self.distributed_descriptor.zeros(dtype=self.dtype)
        Bsc_mm = self.distributed_descriptor.zeros(dtype=self.dtype)
        vec_mm = self.distributed_descriptor.zeros(dtype=self.dtype)

        temporary_eps = np.zeros([self.arraysize])
        if is_master:
            assert self.blacsgrid.comm.rank == 0
            Asc_MM[:, :] = A
            Bsc_MM[:, :] = B

        self.head_to_all_redistributor.redistribute(Asc_MM, Asc_mm)
        self.head_to_all_redistributor.redistribute(Bsc_MM, Bsc_mm)

        self.distributed_descriptor.general_diagonalize_dc(
            Asc_mm, Bsc_mm, vec_mm, temporary_eps)

        # vec_MM contains the eigenvectors in 'Fortran form'. They need to be
        # transpose-conjugated before they are consistent with Scipy behaviour
        self.all_to_head_redistributor.redistribute(vec_mm, vec_MM, uplo="G")

        if is_master:
            # Conjugate-transpose here since general_diagonalize_dc gives us
            # Fortran-convention eigenvectors.
            A[:, :] = vec_MM.conj().T
            eps[:] = temporary_eps



import pytest
import numpy as np
from ase import Atoms
from ase.units import Ha, Bohr

from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.grid_descriptor import GridDescriptor
from gpaw.kpoint import KPoint
from gpaw.symmetry import Symmetry
from gpaw.wavefunctions.arrays import PlaneWaveExpansionWaveFunctions
from gpaw.pw.descriptor import PWDescriptor
from gpaw.pw.lfc import PWLFC
from gpaw.projections import Projections
from gpaw.mpi import world
from gpaw.spline import Spline
from gpaw.hybrids.energy import calculate_energy
from gpaw.hybrids.forces import calculate_forces
from gpaw.hybrids.coulomb import coulomb_interaction
from gpaw.hybrids.paw import calculate_paw_stuff
from gpaw.hybrids.symmetry import Symmetry as Sym
from gpaw.hybrids.kpts import get_kpt


N = 20
L = 2.5
nb = 2
r2 = np.linspace(0, 2, 101)**2


@pytest.mark.ci
@pytest.mark.skipif(world.size > 1, reason='Not parallelized')
def test_force():
    x = 0.2
    y = 0.35
    z = 0.1
    calc = Calc([[x, y, z]])
    omega = 0.2
    wfs = calc.wfs
    kd = wfs.kd
    coulomb = coulomb_interaction(omega, wfs.gd, kd)
    sym = Sym(kd)
    paw_s = calculate_paw_stuff(wfs, calc.density)
    F_v = calculate_forces(wfs, coulomb, sym, paw_s)[0] * Ha / Bohr
    print(F_v)
    dx = 0.0001
    y -= dx / 2
    calc = Calc([[x, y, z]])
    em = energy(calc, sym, coulomb)
    y += dx
    calc = Calc([[x, y, z]])
    ep = energy(calc, sym, coulomb)
    error = (em - ep) / (dx * L * Bohr) - F_v[1]
    print(error)
    assert abs(error) < 6e-10


class Calc:
    def __init__(self, spos_ac):
        self.spos_ac = np.array(spos_ac)
        self.wfs = WFS(self.spos_ac)
        self.density = Dens(self.wfs)
        self.hamiltonian = Ham()


class WFS:
    nspins = 1
    world = world

    def __init__(self, spos_ac):
        self.spos_ac = spos_ac
        self.gd = GridDescriptor([N, N, N], np.eye(3) * L)
        sym = Symmetry([], self.gd.cell_cv)
        self.kd = KPointDescriptor(None)
        self.kd.set_symmetry(Atoms(pbc=True), sym)
        self.setups = [Setup()]
        pd = PWDescriptor(50, self.gd, complex, self.kd)
        data = pd.zeros(nb)
        R = self.gd.get_grid_point_coordinates()
        R -= L / 2
        R2 = (R**2).sum(0)
        data[0] = pd.fft(np.exp(-R2 * (10 / L**2)))
        data[1] = pd.fft(np.exp(-R2 * (10 / L**2))) + 0.5
        psit = PlaneWaveExpansionWaveFunctions(nb, pd, data=data)
        proj = Projections(nb, [1], AP(), world, dtype=complex)
        self.pt = PWLFC([[Spline(0, 2.0, np.exp(-r2 * 10))]], pd)
        self.pt.set_positions(self.spos_ac)
        psit.matrix_elements(self.pt, out=proj)
        f_n = np.array([1.0, 0.5])
        kpt = KPoint(1.0, 1.0, 0, 0, 0, None)
        kpt.psit = psit
        kpt.projections = proj
        kpt.f_n = f_n
        self.kpt_u = [kpt]
        self.kpt_qs = [[kpt]]


class Ham:
    e_total_free = -100.0
    e_xc = -20.0


class Dens:
    nspins = 1

    def __init__(self, wfs):
        kpt = wfs.kpt_u[0]
        self.D_asp = D(kpt)
        self.finegd = wfs.gd.refine()
        self.nt_sg = self.finegd.zeros(1)


class D(dict):
    def __init__(self, kpt):
        dict.__init__(self)
        P_ni = kpt.P_ani[0]
        self[0] = kpt.f_n.dot(P_ni * P_ni.conj()).real[np.newaxis]
        self.partition = AP()


class AP:
    my_indices = [0]
    comm = world
    rank_a = [0]


class Setup:
    Delta_iiL = np.zeros((1, 1, 1)) + 0.1
    X_p = np.zeros(1) + 0.3
    ExxC = -10.0
    ghat_l = [Spline(0, 2.0, np.exp(-r2 * 10))]
    xc_correction = None
    M_pp = np.zeros((1, 1)) + 0.3


def energy(calc, sym, coulomb):
    wfs = calc.wfs
    paw_s = calculate_paw_stuff(wfs, calc.density)
    kpts = [get_kpt(wfs, 0, 0, 0, nb)]
    e1, e2 = calculate_energy(kpts, paw_s[0], wfs, sym, coulomb, wfs.spos_ac)
    evc = e1 * 2 * Ha
    evv = e2 * 2 * Ha
    return evc + evv



import numbers
import tarfile
import xml.sax

import numpy as np

from gpaw.mpi import broadcast as mpi_broadcast
from gpaw.mpi import world

intsize = 4
floatsize = np.array([1], float).itemsize
complexsize = np.array([1], complex).itemsize
itemsizes = {'int': intsize, 'float': floatsize, 'complex': complexsize}

    
class FileReference:
    """Common base class for having reference to a file. The actual I/O
       classes implementing the referencing should be inherited from
       this class."""

    def __init__(self):
        raise NotImplementedError('Should be implemented in derived classes')

    def __len__(self):
        raise NotImplementedError('Should be implemented in derived classes')

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def __getitem__(self):
        raise NotImplementedError('Should be implemented in derived classes')

    def __array__(self):
        return self[::]


def open(filename, mode='r', comm=world):
    if not filename.endswith('.gpw'):
        filename += '.gpw'

    assert mode == 'r'
    return Reader(filename, comm)


class Reader(xml.sax.handler.ContentHandler):
    def __init__(self, name, comm=world):
        self.comm = comm  # used for broadcasting replicated data
        self.master = (self.comm.rank == 0)
        self.dims = {}
        self.shapes = {}
        self.dtypes = {}
        self.parameters = {}
        xml.sax.handler.ContentHandler.__init__(self)
        self.tar = tarfile.open(name, 'r')
        f = self.tar.extractfile('info.xml')
        xml.sax.parse(f, self)

    def startElement(self, tag, attrs):
        if tag == 'gpaw_io':
            self.byteswap = ((attrs['endianness'] == 'little') !=
                             np.little_endian)
        elif tag == 'array':
            name = attrs['name']
            self.dtypes[name] = attrs['type']
            self.shapes[name] = []
            self.name = name
        elif tag == 'dimension':
            n = int(attrs['length'])
            self.shapes[self.name].append(n)
            self.dims[attrs['name']] = n
        else:
            assert tag == 'parameter'
            try:
                value = eval(attrs['value'], {})
            except (SyntaxError, NameError):
                value = str(attrs['value'])
            self.parameters[attrs['name']] = value

    def dimension(self, name):
        return self.dims[name]
    
    def __getitem__(self, name):
        return self.parameters[name]

    def has_array(self, name):
        return name in self.shapes
    
    def get(self, name, *indices, **kwargs):
        broadcast = kwargs.pop('broadcast', False)
        if self.master or not broadcast:
            fileobj, shape, size, dtype = self.get_file_object(name, indices)
            array = np.fromstring(fileobj.read(size), dtype)
            if self.byteswap:
                array = array.byteswap()
            if dtype == np.int32:
                array = np.asarray(array, int)
            array.shape = shape
            if shape == ():
                array = array.item()
        else:
            array = None

        if broadcast:
            array = mpi_broadcast(array, 0, self.comm)
        return array
    
    def get_reference(self, name, indices, length=None):
        fileobj, shape, size, dtype = self.get_file_object(name, indices)
        assert dtype != np.int32
        return TarFileReference(fileobj, shape, dtype, self.byteswap, length)
    
    def get_file_object(self, name, indices):
        dtype, type, itemsize = self.get_data_type(name)
        fileobj = self.tar.extractfile(name)
        n = len(indices)
        shape = self.shapes[name]
        size = itemsize * np.prod(shape[n:], dtype=int)
        offset = 0
        stride = size
        for i in range(n - 1, -1, -1):
            offset += indices[i] * stride
            stride *= shape[i]
        fileobj.seek(offset)
        return fileobj, shape[n:], size, dtype

    def get_data_type(self, name):
        type = self.dtypes[name]
        dtype = np.dtype({'int': np.int32,
                          'float': float,
                          'complex': complex}[type])
        return dtype, type, dtype.itemsize

    def get_parameters(self):
        return self.parameters

    def close(self):
        self.tar.close()


class TarFileReference(FileReference):
    def __init__(self, fileobj, shape, dtype, byteswap, length):
        self.fileobj = fileobj
        self.shape = tuple(shape)
        self.dtype = dtype
        self.itemsize = dtype.itemsize
        self.byteswap = byteswap
        self.offset = fileobj.tell()
        self.length = length

    def __len__(self):
        return self.shape[0]

    def __getitem__(self, indices):
        if isinstance(indices, slice):
            start, stop, step = indices.indices(len(self))
            if start != 0 or step != 1 or stop != len(self):
                raise NotImplementedError('You can only slice a TarReference '
                                          'with [:] or [int]')
            else:
                indices = ()
        elif isinstance(indices, numbers.Integral):
            indices = (indices,)
        else:  # Probably tuple or ellipsis
            raise NotImplementedError('You can only slice a TarReference '
                                      'with [:] or [int]')
            
        n = len(indices)

        size = np.prod(self.shape[n:], dtype=int) * self.itemsize
        offset = self.offset
        stride = size
        for i in range(n - 1, -1, -1):
            offset += indices[i] * stride
            stride *= self.shape[i]
        self.fileobj.seek(offset)
        array = np.fromstring(self.fileobj.read(size), self.dtype)
        if self.byteswap:
            array = array.byteswap()
        array.shape = self.shape[n:]
        if self.length:
            array = array[..., :self.length].copy()
        return array



from time import time

import numpy as np
from scipy.linalg import eigh

# Set-up a simple matrix in parallel, diagonalize using ScaLAPACK
# D&C driver then compare *eigenvalues* with serial LAPACK diagonlize

from gpaw.blacs import BlacsGrid
from gpaw.mpi import world
from gpaw.utilities.scalapack import scalapack_set, \
    scalapack_zero


switch_uplo = {'U': 'L', 'L': 'U'}
rank = world.rank
tol = 5e-12  # eigenvalue tolerance


def test_parallel_scalapack_diag_simple(scalapack):
    nbands = 1000
    mb = 64

    if world.size == 1:
        blacsshape = (1, 1)
    else:
        blacsshape = (world.size // 2, 2)

    # Set-up BlacsGrud
    grid = BlacsGrid(world, *blacsshape)

    # Create descriptor
    nndesc = grid.new_descriptor(nbands, nbands, mb, mb)
    H_nn = nndesc.empty(dtype=float)  # outside BlacsGrid these are size zero
    C_nn = nndesc.empty(dtype=float)  # outside BlacsGrid these are size zero
    eps_N = np.empty((nbands), dtype=float)  # replicated on all MPI tasks
    # Fill ScaLAPACK array
    alpha = 0.1  # off-diagonal
    beta = 75.0  # diagonal
    uplo = 'L'  # lower-triangular
    scalapack_set(nndesc, H_nn, alpha, beta, uplo)
    scalapack_zero(nndesc, H_nn, switch_uplo[uplo])

    t1 = time()
    # either interface will work, we recommend use the latter interface
    # scalapack_diagonalize_dc(nndesc, H_nn.copy(), C_nn, eps_N, 'L')
    nndesc.diagonalize_dc(H_nn.copy(), C_nn, eps_N)
    t2 = time()
    world.broadcast(eps_N, 0)  # all MPI tasks now have eps_N
    world.barrier()  # wait for everyone to finish

    if rank == 0:
        print('ScaLAPACK diagonalize_dc', t2 - t1)

    # Create replicated NumPy array
    diagonal = np.eye(nbands, dtype=float)
    offdiagonal = np.tril(np.ones((nbands, nbands)), -1)
    H0 = beta * diagonal + alpha * offdiagonal
    E0 = np.empty((nbands), dtype=float)

    t1 = time()
    E0 = eigh(H0, eigvals_only=True)
    t2 = time()

    if rank == 0:
        print('LAPACK diagonalize', t2 - t1)

    delta = abs(E0 - eps_N).max()
    if rank == 0:
        print(delta)
        assert delta < tol



from ase.data import atomic_numbers, chemical_symbols
from ase.units import Bohr

from gpaw.setup import Setups
from gpaw.xc import XC
from gpaw.mpi import world

Bondi64jpc_vdWradii = {  # units Anstrom
    'He': 1.40,
    'Ne': 1.54,
    'Ar': 1.88,
    'Kr': 2.02,
    'Xe': 2.16
}
# Van der Waals Radii after
# Pekka Pyykko, Chem. Rev. 97 (1997) 597-636
# Table 2
Pyykko97cr_vdWradii = {  # units Anstrom
    'Ne': 1.55,
    'Ar': 1.88,
    'Kr': 2.00,
    'Xe': 2.18,
    'Rn': 2.24
}
collected_vdWradii = Bondi64jpc_vdWradii
collected_vdWradii['Rn'] = Pyykko97cr_vdWradii['Rn']


def vdWradii(symbols, xc):
    """Find the elements van der Waals radius.

    Method proposed in:
    Tkatchenko and Scheffler PRL 102 (2009) 073005

    The returned radii are given in Angstroms.
    """
    Z_rare_gas = [atomic_numbers[symbol] for symbol in Bondi64jpc_vdWradii]
    Z_rare_gas.append(atomic_numbers['Rn'])
    Z_rare_gas.sort()

    if isinstance(xc, str):
        xc = XC(xc)

    def get_density(Z):
        """Return density and radial grid from setup."""
        # load setup
        setups = Setups([Z], 'paw', {}, xc, world=world)
        setup = setups[0].data
        #  create density
        n_g = setup.nc_g.copy()
        for f, phi_g in zip(setup.f_j, setup.phi_jg):
            n_g += f * phi_g ** 2
        return n_g, setup.rgd.r_g

    radii = []
    radius = {}
    for symbol in symbols:
        Z = atomic_numbers[symbol]
        if symbol not in radius:
            # find the rare gas of the elements row
            Zrg = None
            for Zr in Z_rare_gas:
                if Zrg is None and Z <= Zr:
                    Zrg = Zr

            n_g, r_g = get_density(Zrg)
            # find density at R
            R = collected_vdWradii[chemical_symbols[Zrg]] / Bohr
            n = 0
            while r_g[n] < R:
                n += 1
            # linear interpolation
            ncut = (n_g[n - 1] +
                    (n_g[n] - n_g[n - 1]) * (R - r_g[n - 1]) /
                    (r_g[n] - r_g[n - 1]))
#            print "Z, Zrg, ncut", Z, Zrg, ncut

            # find own R at this density
            n_g, r_g = get_density(Z)
            n = 0
            while n_g[n] > ncut:
                n += 1
            # linear interpolation
            R = (r_g[n - 1] +
                 (r_g[n] - r_g[n - 1]) * (ncut - n_g[n - 1]) /
                 (n_g[n] - n_g[n - 1]))
            radius[symbol] = R * Bohr

        radii.append(radius[symbol])

    return radii



import time
import pytest
import numpy as np

from ase import Atoms
from ase.parallel import parprint

from gpaw import GPAW, PW
from gpaw.test import findpeak
from gpaw.bztools import find_high_symmetry_monkhorst_pack
from gpaw.response.df import DielectricFunction, read_response_function
from gpaw.mpi import size, world


@pytest.mark.response
def test_response_Na_EELS_RPA_tetra_point_comparison(in_tmp_dir):
    assert size <= 4**3

    # Ground state calculation
    t1 = time.time()

    a = 3
    atom = Atoms('Na',
                 cell=[a, 0, 0],
                 pbc=[1, 1, 1])
    atom.center(vacuum=1 * a, axis=(1, 2))
    atom.center()
    atoms = atom.repeat((2, 1, 1))

    calc = GPAW(mode=PW(200),
                nbands=4,
                setups={'Na': '1'},
                kpts=(16, 2, 2),
                parallel={'band': 1},
                xc='LDA')

    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write('Na_gs')

    # Generate grid compatible with tetrahedron integration
    kpts = find_high_symmetry_monkhorst_pack('Na_gs', 6.0)

    # Calculate the wave functions on the new kpts grid
    calc = GPAW('Na_gs').fixed_density(kpts=kpts)
    calc.write('Na', 'all')

    t2 = time.time()

    # Excited state calculation
    q0_c = np.array([0., 0., 0.])
    q1_c = np.array([1 / 4., 0., 0.])
    w_w = np.linspace(0, 24, 241)

    # Calculate the eels spectrum using point integration at both q-points
    df1 = DielectricFunction(calc='Na', frequencies=w_w, eta=0.2, ecut=50,
                             hilbert=False, rate=0.2)
    df1.get_eels_spectrum(xc='RPA', filename='EELS_Na-PI_q0', q_c=q0_c)
    df1.get_eels_spectrum(xc='RPA', filename='EELS_Na-PI_q1', q_c=q1_c)

    t3 = time.time()

    # Calculate the eels spectrum using tetrahedron integration at both q
    df2 = DielectricFunction(calc='Na', eta=0.2, ecut=50,
                             integrationmode='tetrahedron integration',
                             hilbert=True, rate=0.2)
    df2.get_eels_spectrum(xc='RPA', filename='EELS_Na-TI_q0', q_c=q0_c)
    df2.get_eels_spectrum(xc='RPA', filename='EELS_Na-TI_q1', q_c=q1_c)

    t4 = time.time()

    parprint('')
    parprint('For ground  state calc, it took', (t2 - t1) / 60, 'minutes')
    parprint('For PI excited state calc, it took', (t3 - t2) / 60, 'minutes')
    parprint('For TI excited state calc, it took', (t4 - t3) / 60, 'minutes')

    world.barrier()
    omegaP0_w, _, eelsP0_w = read_response_function('EELS_Na-PI_q0')
    omegaP1_w, _, eelsP1_w = read_response_function('EELS_Na-PI_q1')
    omegaT0_w, _, eelsT0_w = read_response_function('EELS_Na-TI_q0')
    omegaT1_w, _, eelsT1_w = read_response_function('EELS_Na-TI_q1')

    # calculate tetra & point int. peaks and intensities
    wpeakP0, IpeakP0 = findpeak(omegaP0_w, eelsP0_w)
    wpeakP1, IpeakP1 = findpeak(omegaP1_w, eelsP1_w)
    wpeakT0, IpeakT0 = findpeak(omegaT0_w, eelsT0_w)
    wpeakT1, IpeakT1 = findpeak(omegaT1_w, eelsT1_w)

    # import matplotlib.pyplot as plt
    # plt.subplot(1, 2, 1)
    # plt.plot(omegaP0_w, eelsP0_w)
    # plt.plot(omegaT0_w, eelsT0_w)
    # plt.subplot(1, 2, 2)
    # plt.plot(omegaP1_w, eelsP1_w)
    # plt.plot(omegaT1_w, eelsT1_w)
    # plt.show()

    # tetra and point integrators should produce similar results:
    # confirm this by comparing the 2 integration methods
    assert wpeakT0 == pytest.approx(wpeakP0, abs=0.05)
    assert wpeakT1 == pytest.approx(wpeakP1, abs=0.1)

    # ensure the wpeak for point & tetra integration do not change
    assert wpeakP0 == pytest.approx(3.4811, abs=0.02)
    assert wpeakP1 == pytest.approx(3.8076, abs=0.02)
    assert wpeakT0 == pytest.approx(3.5033, abs=0.02)
    assert wpeakT1 == pytest.approx(3.8230, abs=0.11)

    # ensure the Ipeak for point & tetra integration do not change
    assert IpeakP0 == pytest.approx(8.6311, abs=1.)
    assert IpeakP1 == pytest.approx(7.7766, abs=1.)
    assert IpeakT0 == pytest.approx(8.3715, abs=1.)
    assert IpeakT1 == pytest.approx(7.7335, abs=1.)



from typing import Any, Dict

import numpy as np
from ase.units import Ha

from gpaw.density import Density
from gpaw.external import NoExternalPotential
from gpaw.pw.hamiltonian import ReciprocalSpaceHamiltonian
from gpaw.typing import Array1D, Array2D, ArrayLike1D


class BField(NoExternalPotential):
    def __init__(self, field: ArrayLike1D):
        """Constant magnetic field.

        field:
            B-field vector in units of eV/bohr-magnoton.
        """
        self.name = 'BField'
        self.field_v = np.array(field) / Ha
        assert self.field_v.shape == (3,)

    def get_potential(self, gd):
        raise NotImplementedError('BField can only be used in PW-mode!')

    def update_potential_pw(self,
                            ham: ReciprocalSpaceHamiltonian,
                            dens: Density) -> float:
        magmom_v, _ = dens.estimate_magnetic_moments()
        eext = -self.field_v.dot(magmom_v)
        if dens.collinear:
            ham.vt_sG[:] = ham.pd2.ifft(ham.vt_Q)
            ham.vt_sG[0] -= self.field_v[2]
            ham.vt_sG[1] += self.field_v[2]
        else:
            ham.vt_xG[0] = ham.pd2.ifft(ham.vt_Q)
            ham.vt_xG[1:] = -self.field_v.reshape((3, 1, 1, 1))
        return eext

    def paw_correction(self, Delta_p: Array1D, dH_sp: Array2D) -> None:
        if len(dH_sp) == 2:
            c = (4 * np.pi)**0.5 * self.field_v[2]
            dH_sp[0] -= c * Delta_p
            dH_sp[1] += c * Delta_p
        else:
            c_vp = (4 * np.pi)**0.5 * self.field_v[:, np.newaxis]
            dH_sp[1:] -= c_vp * Delta_p

    def todict(self) -> Dict[str, Any]:
        return {'name': self.name,
                'field': tuple(self.field_v * Ha)}



import numpy as np

from gpaw.test import equal
from gpaw.lrtddft.apmb import sqrt_matrix

# check sqrt of a matrix


def test_linalg_lapack():
    A = [[20, 4], [4, 1]]
    a = [[4.4, 0.8], [0.8, 0.6]]
    A = np.array(A, float)
    print('A=', A)
    a = np.array(a)
    b = sqrt_matrix(A)
    print('sqrt(A)=', b)
    equal(((a - b)**2).sum(), 0, 1.0e-12)



from gpaw.test import equal
from gpaw.cluster import Cluster
from ase.build import molecule
from ase.units import Pascal, m
from ase.data.vdw import vdw_radii
from gpaw.solvation import (
    SolvationGPAW,
    EffectivePotentialCavity,
    Power12Potential,
    LinearDielectric,
    SurfaceInteraction,
    VolumeInteraction,
    LeakedDensityInteraction,
    GradientSurface,
    KB51Volume)
import numpy as np

vdw_radii = vdw_radii.copy()


def test_solvation_spinpol():
    h = 0.3
    vac = 3.0
    u0 = .180
    epsinf = 80.
    T = 298.15
    vdw_radii[1] = 1.09

    def atomic_radii(atoms):
        return [vdw_radii[n] for n in atoms.numbers]

    atoms = Cluster(molecule('CN'))
    atoms.minimal_box(vac, h)
    atoms2 = atoms.copy()
    atoms2.set_initial_magnetic_moments(None)

    atomss = (atoms, atoms2)
    Es = []
    Fs = []

    for atoms in atomss:
        atoms.calc = SolvationGPAW(
            mode='fd', xc='LDA', h=h, charge=-1,
            cavity=EffectivePotentialCavity(
                effective_potential=Power12Potential(atomic_radii, u0),
                temperature=T,
                surface_calculator=GradientSurface(),
                volume_calculator=KB51Volume()
            ),
            dielectric=LinearDielectric(epsinf=epsinf),
            interactions=[
                SurfaceInteraction(
                    surface_tension=100. * 1e-3 * Pascal * m
                ),
                VolumeInteraction(
                    pressure=-1.0 * 1e9 * Pascal
                ),
                LeakedDensityInteraction(
                    voltage=1.0
                )
            ]
        )
        Es.append(atoms.get_potential_energy())
        Fs.append(atoms.get_forces())

    # compare to expected difference of a gas phase calc
    print('difference E: ', Es[0] - Es[1])
    equal(Es[0], Es[1], 0.0002)
    print('difference F: ', np.abs(Fs[0] - Fs[1]).max())
    equal(Fs[0], Fs[1], 0.003)

    # XXX add test case where spin matters, e.g. charge=0 for CN?



import sys
import time

import numpy as np

import ase.db
from ase.units import Rydberg
from ase.utils import opencew
from ase.calculators.calculator import kpts2mp
from ase.io import Trajectory
from ase.test.tasks.dcdft import DeltaCodesDFTCollection as Collection
from gpaw import GPAW, PW, FermiDirac
from gpaw.utilities import h2gpts

collection = Collection()

if len(sys.argv) == 1:
    names = collection.names
else:
    names = [sys.argv[1]]

c = ase.db.connect('dcdft_gpaw_pw.db')

# mode = 'lcao'
mode = 'fd'
mode = 'pw'

e = 0.08  # h -> gpts
e = round(100 * Rydberg, 0)

kptdensity = 16.0  # this is converged
kptdensity = 6.0  # just for testing
width = 0.01

relativistic = True
constant_basis = True

if relativistic:
    linspace = (0.98, 1.02, 5)  # eos numpy's linspace
else:
    linspace = (0.92, 1.08, 7)  # eos numpy's linspace
linspacestr = ''.join([str(t) + 'x' for t in linspace])[:-1]

code = ('gpaw' + '-' + mode + str(e) + '_c' +
        str(constant_basis) + '_e' + linspacestr)
code = code + '_k' + str(kptdensity) + '_w' + str(width)
code = code + '_r' + str(relativistic)

for name in names:
    # save all steps in one traj file in addition to the database
    # we should only used the database c.reserve, but here
    # traj file is used as another lock ...
    fd = opencew(name + '_' + code + '.traj')
    if fd is None:
        continue
    traj = Trajectory(name + '_' + code + '.traj', 'w')
    atoms = collection[name]
    cell = atoms.get_cell()
    kpts = tuple(kpts2mp(atoms, kptdensity, even=True))
    kwargs = {}
    if mode in ['fd', 'lcao']:
        if constant_basis:
            # gives more smooth EOS in fd mode
            kwargs.update({'gpts': h2gpts(e, cell)})
        else:
            kwargs.update({'h': e})
    elif mode == 'pw':
        if constant_basis:
            kwargs.update({'mode': PW(e, cell=cell)})
            kwargs.update({'gpts': h2gpts(0.10, cell)})
        else:
            kwargs.update({'mode': PW(e)})
    if mode == 'pw':
        if name in ['Li', 'Na']:
            # listserv.fysik.dtu.dk/pipermail/gpaw-developers/2012-May/002870.html
            if constant_basis:
                kwargs.update({'gpts': h2gpts(0.05, cell)})
            else:
                kwargs.update({'h': 0.05})
    if mode == 'lcao':
        kwargs.update({'mode': 'lcao'})
        kwargs.update({'basis': 'dzp'})
    if mode == 'fd':
        kwargs.update({'mode': 'fd'})
    if name in ['He', 'Ne', 'Ar', 'Kr', 'Xe', 'Rn', 'Ca', 'Sr', 'Ba', 'Be']:
        # results wrong / scf slow with minimal basis
        kwargs.update({'basis': 'dzp'})
        kwargs.update({'nbands': -5})
    # loop over EOS linspace
    for n, x in enumerate(np.linspace(linspace[0], linspace[1], linspace[2])):
        id = c.reserve(name=name, mode=mode, e=e, linspacestr=linspacestr,
                       kptdensity=kptdensity, width=width,
                       relativistic=relativistic,
                       constant_basis=constant_basis,
                       x=x)
        if id is None:
            continue
        # perform EOS step
        atoms.set_cell(cell * x, scale_atoms=True)
        # set calculator
        base_kwargs = dict(
            txt=name + '_' + code + '_' + str(n) + '.txt',
            xc='PBE',
            kpts=kpts,
            occupations=FermiDirac(width),
            parallel={'band': 1},
            maxiter=777)
        atoms.calc = GPAW(**dict(base_kwargs, **kwargs))
        t = time.time()
        atoms.get_potential_energy()
        c.write(atoms,
                name=name, mode=mode, e=e, linspacestr=linspacestr,
                kptdensity=kptdensity, width=width,
                relativistic=relativistic,
                constant_basis=constant_basis,
                x=x,
                niter=atoms.calc.get_number_of_iterations(),
                time=time.time() - t)
        traj.write(atoms)
        del c[id]
